"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json

from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CloudFormationLintRule


class EqualsIsUseful(CloudFormationLintRule):
    """
    Validate that the Equals will return
    true/false and not always be true or false
    """

    id = "W8003"
    shortdesc = "Fn::Equals will always return true or false"
    description = (
        "Validate Fn::Equals to see if its comparing two strings or two equal items."
        " While this works it may not be intended."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-equals"
    tags = ["functions", "equals"]

    def equals_is_useful(self, validator, s, instance, schema):
        if not validator.is_type(instance, "array"):
            return

        if len(instance) != 2:
            return

        # testing on 2023/09/24
        # True (boolean) != "True"
        # True (boolean) == "true"
        # 1 == "1"
        if json.dumps(instance[0]) == json.dumps(instance[1]):
            yield ValidationError(
                f"{instance!r} will always return {True!r} or {False!r}",
                rule=self,
            )
            return
        try:
            first = instance[0]
            second = instance[1]
            if validator.is_type(first, "boolean"):
                first = "true" if first else "false"
            if validator.is_type(second, "boolean"):
                first = "true" if first else "false"
            if str(first) == str(second):
                yield ValidationError(
                    f"{instance!r} will always return {True!r}",
                    rule=self,
                )
            if isinstance(instance[0], (str, float, int, bool)) and isinstance(
                instance[1], (str, float, int, bool)
            ):
                if str(first) != str(second):
                    yield ValidationError(
                        f"{instance!r} will always return {False!r}",
                        rule=self,
                    )
        except:  # noqa: E722
            pass



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.rules.RuleCondition import RuleCondition


@pytest.fixture(scope="module")
def rule():
    rule = RuleCondition()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "boolean is okay",
            True,
            [],
        ),
        (
            "wrong type",
            [],
            [
                ValidationError(
                    "[] is not of type 'boolean'",
                    validator="type",
                    schema_path=deque(["type"]),
                    rule=RuleCondition(),
                )
            ],
        ),
        (
            "functions are okay",
            {"Fn::Equals": ["a", "b"]},
            [],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.Select import Select
from cfnlint.rules.functions.SelectResolved import SelectResolved


@pytest.fixture(scope="module")
def rule():
    rule = Select()
    rule.child_rules["W1035"] = SelectResolved()
    yield rule


@pytest.fixture
def template():
    return {
        "Mappings": {
            "A": {
                "B": {
                    "C": "D",
                },
            },
        }
    }


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Select with array",
            {"Fn::Select": [1, ["bar"]]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Select is NOT a array",
            {"Fn::Select": "foo"},
            {"type": "string"},
            [
                ValidationError(
                    "'foo' is not of type 'array'",
                    path=deque(["Fn::Select"]),
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    validator="fn_select",
                ),
            ],
        ),
        (
            "Invalid Fn::Length using an invalid function for index",
            {"Fn::Select": [{"Fn::Length": [1, 2]}, ["bar"]]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::Length': [1, 2]} is not of type 'integer'",
                    path=deque(["Fn::Select", 0]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "else",
                            "prefixItems",
                            0,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_select",
                ),
            ],
        ),
        (
            "Invalid Fn::Select using an invalid function for array",
            {"Fn::Select": [1, {"foo": "bar"}]},
            {"type": "string"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'array'",
                    path=deque(["Fn::Select", 1]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "else",
                            "prefixItems",
                            1,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_select",
                ),
            ],
        ),
        (
            "Invalid Fn::Select using an invalid CFN function",
            {"Fn::Select": [1, {"Fn::Join": ["-", "bar"]}]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::Join': ['-', 'bar']} is not of type 'array'",
                    path=deque(["Fn::Select", 1]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "else",
                            "prefixItems",
                            1,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_select",
                ),
            ],
        ),
        (
            "Valid Fn::Select with a valid function",
            {"Fn::Select": [1, {"Fn::Split": ["-", "bar"]}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Select with a valid function (FindInMap)",
            {"Fn::Select": [1, ["foo", {"Fn::FindInMap": ["A", "B", "C"]}]]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Select with an invalid function",
            {"Fn::Select": [1, ["foo", {"foo": "bar"}]]},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "{'Fn::Select': [1, ['foo', {'foo': 'bar'}]]} is not of type "
                        "'string' when 'Fn::Select' is resolved"
                    ),
                    path=deque(["Fn::Select"]),
                    schema_path=deque(["type"]),
                    validator="fn_select",
                    rule=SelectResolved(),
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_select(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.parameters.Enum import Enum as ParameterEnum
from cfnlint.rules.resources.properties.Enum import Enum


@pytest.fixture
def rule():
    rule = Enum()
    rule.child_rules["W2030"] = ParameterEnum()
    yield Enum()


@pytest.fixture
def template():
    return {
        "Parameters": {
            "MyString": {
                "Type": "String",
                "Default": "a",
                "AllowedValues": ["a", "b"],
            },
        },
        "Resources": {},
    }


@pytest.mark.parametrize(
    "name,instance,enums,path,expected",
    [
        (
            "Validate the standard enum",
            "a",
            ["a"],
            {},
            [],
        ),
        (
            "Validate the standard enum",
            "a",
            ["a"],
            {"value_path": ["Parameters", "MyString", "AllowedValues", 0]},
            [],
        ),
        (
            "Invalid with the standard enum",
            "a",
            ["b"],
            {},
            [
                ValidationError(
                    ("'a' is not one of ['b']"),
                    path=deque([]),
                    schema_path=deque([]),
                    path_override=deque([]),
                ),
            ],
        ),
        (
            "Invalid with the standard enum from parameter",
            "a",
            ["b"],
            {"value_path": ["Parameters", "MyString", "AllowedValues", 0]},
            [
                ValidationError(
                    ("'a' is not one of ['b']"),
                    path=deque([]),
                    schema_path=deque([]),
                    rule=ParameterEnum(),
                    path_override=deque(["Parameters", "MyString", "AllowedValues", 0]),
                ),
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate_enum(name, instance, enums, expected, rule, validator):
    errs = list(rule.enum(validator, enums, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"


@pytest.mark.parametrize(
    "name,instance,enums,path,expected",
    [
        (
            "Validate the standard enum",
            "a",
            ["A"],
            {},
            [],
        ),
        (
            "Validate the standard enum",
            "a",
            ["A"],
            {"value_path": ["Parameters", "MyString", "AllowedValues", 0]},
            [],
        ),
        (
            "Invalid with the standard enum",
            "a",
            ["B"],
            {},
            [
                ValidationError(
                    ("'a' is not one of ['b'] (case-insensitive)"),
                    path=deque([]),
                    schema_path=deque([]),
                    path_override=deque([]),
                ),
            ],
        ),
        (
            "Invalid with the standard enum from parameter",
            "a",
            ["B"],
            {"value_path": ["Parameters", "MyString", "AllowedValues", 0]},
            [
                ValidationError(
                    ("'a' is not one of ['b'] (case-insensitive)"),
                    path=deque([]),
                    schema_path=deque([]),
                    rule=ParameterEnum(),
                    path_override=deque(["Parameters", "MyString", "AllowedValues", 0]),
                ),
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate_enum_case_insensitive(
    name, instance, enums, expected, rule, validator
):
    errs = list(rule.enumCaseInsensitive(validator, enums, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema


class CreationPolicy(CfnLintJsonSchema):
    id = "E3055"
    shortdesc = "Check CreationPolicy values for Resources"
    description = "Check that the CreationPolicy values are valid"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-creationpolicy.html"
    tags = ["resources", "creationPolicy"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/*/CreationPolicy"],
            all_matches=True,
        )

    def _get_schema(self, resource_type: str) -> dict[str, Any]:
        if resource_type == "AWS::AppStream::Fleet":
            return {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "StartFleet": {
                        "additionalProperties": False,
                        "type": "object",
                        "properties": {"Type": {"type": "boolean"}},
                    }
                },
            }
        if resource_type == "AWS::AutoScaling::AutoScalingGroup":
            return {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "AutoScalingCreationPolicy": {
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "MinSuccessfulInstancesPercent": {"type": "integer"}
                        },
                    },
                    "ResourceSignal": {
                        "additionalProperties": False,
                        "type": "object",
                        "properties": {
                            "Timeout": {"type": "string"},
                            "Count": {"type": "integer"},
                        },
                    },
                },
            }
        if resource_type == "AWS::CloudFormation::WaitCondition":

            return {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "ResourceSignal": {
                        "additionalProperties": False,
                        "type": "object",
                        "properties": {
                            "Timeout": {"type": "string"},
                            "Count": {"type": "integer"},
                        },
                    }
                },
            }

        return {}

    # pylint: disable=unused-argument, arguments-renamed
    def validate(self, validator: Validator, dP: str, instance, schema):
        resource_name = validator.context.path.path[1]
        if not isinstance(resource_name, str):
            return
        resource_type = validator.context.resources[resource_name].type

        validator = validator.evolve(
            context=validator.context.evolve(
                functions=[
                    "Fn::Sub",
                    "Fn::Select",
                    "Fn::FindInMap",
                    "Fn::If",
                    "Ref",
                ],
                strict_types=False,
            ),
            schema=self._get_schema(resource_type),
        )

        yield from self._iter_errors(validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.SecurityGroupAllToAndFromPorts import (
    SecurityGroupAllToAndFromPorts,
)


@pytest.fixture(scope="module")
def rule():
    rule = SecurityGroupAllToAndFromPorts()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "ToPort": -1,
                "FromPort": -1,
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "ToPort": "-1",
                "FromPort": -1,
            },
            [],
        ),
        (
            {
                "IpProtocol": 1,
                "ToPort": -1,
                "FromPort": 8,
            },
            [],
        ),
        (
            {
                "ToPort": -1,
            },
            [
                ValidationError(
                    ("Both ['FromPort', 'ToPort'] must " "be -1 when one is -1"),
                    rule=SecurityGroupAllToAndFromPorts(),
                    path=deque([]),
                    validator="required",
                    validator_value=["FromPort"],
                    instance={"ToPort": -1},
                    schema={
                        "properties": {"FromPort": {"enum": [-1, "-1"]}},
                        "required": ["FromPort"],
                    },
                    schema_path=deque(["else", "allOf", 0, "then", "required"]),
                )
            ],
        ),
        (
            {
                "FromPort": -1,
            },
            [
                ValidationError(
                    ("Both ['FromPort', 'ToPort'] must " "be -1 when one is -1"),
                    rule=SecurityGroupAllToAndFromPorts(),
                    path=deque([]),
                    validator="required",
                    validator_value=["ToPort"],
                    instance={"FromPort": -1},
                    schema={
                        "properties": {"ToPort": {"enum": [-1, "-1"]}},
                        "required": ["ToPort"],
                    },
                    schema_path=deque(["else", "allOf", 1, "then", "required"]),
                )
            ],
        ),
        (
            {
                "ToPort": -1,
                "FromPort": 5,
            },
            [
                ValidationError(
                    ("Both ['FromPort', 'ToPort'] must " "be -1 when one is -1"),
                    rule=SecurityGroupAllToAndFromPorts(),
                    path=deque(["FromPort"]),
                    validator="enum",
                    validator_value=[-1, "-1"],
                    instance=5,
                    schema={"enum": [-1, "-1"]},
                    schema_path=deque(
                        ["else", "allOf", 0, "then", "properties", "FromPort", "enum"]
                    ),
                )
            ],
        ),
        (
            {
                "ToPort": 5,
                "FromPort": -1,
            },
            [
                ValidationError(
                    ("Both ['FromPort', 'ToPort'] must " "be -1 when one is -1"),
                    rule=SecurityGroupAllToAndFromPorts(),
                    path=deque(["ToPort"]),
                    validator="enum",
                    validator_value=[-1, "-1"],
                    instance=5,
                    schema={"enum": [-1, "-1"]},
                    schema_path=deque(
                        ["else", "allOf", 1, "then", "properties", "ToPort", "enum"]
                    ),
                )
            ],
        ),
        (
            {
                "IpProtocol": "icmp",
                "ToPort": 8,
                "FromPort": -1,
            },
            [
                ValidationError(
                    ("Both ['FromPort', 'ToPort'] must " "be -1 when one is -1"),
                    rule=SecurityGroupAllToAndFromPorts(),
                    path=deque(["ToPort"]),
                    validator="enum",
                    validator_value=[-1, "-1"],
                    instance=5,
                    schema={"enum": [-1, "-1"]},
                    schema_path=deque(["then", "then", "properties", "ToPort", "enum"]),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import hashlib
import uuid
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cfnlint.rules import CloudFormationLintRule, RuleMatch


@dataclass(frozen=True, eq=True)
class Match:
    """
    Represents a match found by a CloudFormationLintRule.

    Attributes:
        linenumber (int): The line number where the match was found.
        columnnumber (int): The column number where the match was found.
        linenumberend (int): The ending line number of the match.
        columnnumberend (int): The ending column number of the match.
        filename (str): The name of the file where the match was found.
        rule (CloudFormationLintRule): The rule that found the match.
        message (str): The message associated with the match.
        id (str): A unique identifier for the match.
        parent_id (str): The ID of the parent match,
            if this match is a child of another.

    Methods:
        create(message: str, filename: str, rule: CloudFormationLintRule,
            linenumber: int = None, columnnumber: int = None,
            linenumberend: int = None, columnnumberend: int = None,
            rulematch_obj: RuleMatch = None, parent_id: str = None) -> Match:
            Factory method to create a new Match instance.

    """

    message: str = field(init=True)
    rule: CloudFormationLintRule = field(init=True)
    filename: str | None = field(init=True, default=None)
    linenumber: int = field(init=True, default=1)
    columnnumber: int = field(init=True, default=1)
    linenumberend: int = field(init=True, default=1)
    columnnumberend: int = field(init=True, default=1)
    id: str = field(init=False)
    parent_id: str | None = field(init=True, default=None)
    rulematch_obj: InitVar[RuleMatch | None] = None

    def __post_init__(self, rulematch_obj):
        hex_string = hashlib.md5(
            f"{self}".encode("UTF-8"), usedforsecurity=False
        ).hexdigest()
        super().__setattr__("id", str(uuid.UUID(hex=hex_string)))

        if rulematch_obj:
            for k, v in vars(rulematch_obj).items():
                if not hasattr(self, k):
                    super().__setattr__(k, v)

    def __repr__(self):
        # use the Posix path to keep things consistent across platforms
        file_str = Path(self.filename).as_posix() + ":" if self.filename else ""
        return f"[{self.rule}] ({self.message}) matched {file_str}{self.linenumber}"

    @classmethod
    def create(
        cls,
        message: str,
        filename: str,
        rule: CloudFormationLintRule,
        linenumber: int | None = None,
        columnnumber: int | None = None,
        linenumberend: int | None = None,
        columnnumberend: int | None = None,
        rulematch_obj: RuleMatch | None = None,
        parent_id: str | None = None,
    ) -> "Match":
        """
        Factory method to create a new Match instance.

        Args:
            message (str): The message associated with the match.
            filename (str): The name of the file where the match was found.
            rule (CloudFormationLintRule): The rule that found the match.
            linenumber (int, optional): The line number where the match
                was found.
            columnnumber (int, optional): The column number where the
                match was found.
            linenumberend (int, optional): The ending line number of
                the match.
            columnnumberend (int, optional): The ending column number of
                the match.
            rulematch_obj (RuleMatch, optional): The RuleMatch object that
                generated this Match.
            parent_id (str, optional): The ID of the parent match, if this
                match is a child of another.

        Returns:
            Match: A new Match instance.
        """

        if columnnumber is None:
            columnnumber = 1
        if columnnumberend is None:
            columnnumberend = columnnumber + 1
        if linenumber is None:
            linenumber = 1
        if linenumberend is None:
            linenumberend = linenumber

        return cls(
            linenumber=linenumber,
            columnnumber=columnnumber,
            linenumberend=linenumberend,
            columnnumberend=columnnumberend,
            filename=filename,
            rule=rule,
            message=message,
            rulematch_obj=rulematch_obj,
            parent_id=parent_id,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.BothUpdateReplacePolicyDeletionPolicyNeeded import (
    UpdateReplacePolicyDeletionPolicy,
)


@pytest.fixture(scope="module")
def rule():
    rule = UpdateReplacePolicyDeletionPolicy()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyResource": {
                "Type": "AWS::S3::Bucket",
            }
        }
    }


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid with nothing specified",
            {},
            deque(["Resources", "MyResource"]),
            [],
        ),
        (
            "Valid both update and replacy policy",
            {
                "Type": "AWS::S3::Bucket",
                "DeletionPolicy": "Retain",
                "UpdateReplacePolicy": "Retain",
            },
            deque(["Resources", "MyResource"]),
            [],
        ),
        (
            "Invalid with just UpdatePolicy",
            {"Type": "AWS::S3::Bucket", "DeletionPolicy": "Retain"},
            deque(["Resources", "MyResource"]),
            [
                ValidationError(
                    (
                        "Both 'UpdateReplacePolicy' and 'DeletionPolicy' are "
                        "needed to protect resource from deletion"
                    ),
                    path=deque([]),
                    rule=UpdateReplacePolicyDeletionPolicy(),
                    schema_path=deque(["anyOf"]),
                    validator="anyOf",
                    context=[],
                )
            ],
        ),
        (
            "Invalid with just UpdateReplacePolicy",
            {"Type": "AWS::S3::Bucket", "UpdateReplacePolicy": "Retain"},
            deque(["Resources", "MyResource"]),
            [
                ValidationError(
                    (
                        "Both 'UpdateReplacePolicy' and 'DeletionPolicy' are "
                        "needed to protect resource from deletion"
                    ),
                    path=deque([]),
                    rule=UpdateReplacePolicyDeletionPolicy(),
                    schema_path=deque(["anyOf"]),
                    validator="anyOf",
                    context=[],
                )
            ],
        ),
    ],
)
def test_validate(name, instance, path, expected, rule, validator):
    context = validator.context
    context = context.evolve(path=Path(path=path))
    validator = validator.evolve(context=context)
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbinstance
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class DbInstanceDbInstanceClassWithEngine(CfnLintJsonSchema):
    id = "E3062"
    shortdesc = "Validates RDS DB Instance Class based on Engine and EngineVersion"
    description = (
        "Validates the RDS DB instance types based on 'Engine' "
        "and 'EngineVersion'. 'EngineVersion' is based on the minor "
        "version."
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::RDS::DBInstance/Properties"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.extensions.aws_rds_dbinstance,
                "db_instance_class.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        validator = validator.evolve(
            context=validator.context.evolve(
                functions=[],
            )
        )

        if validator.is_type(instance.get("Engine"), "string"):
            instance["Engine"] = instance["Engine"].lower()

        yield from super().validate(validator, keywords, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.apigateway.RestApiMixingDefinitions import (
    RestApiMixingDefinitions,
)


@pytest.fixture(scope="module")
def rule():
    rule = RestApiMixingDefinitions()
    yield rule


_rest_api = {
    "Type": "AWS::ApiGateway::RestApi",
    "Properties": {},
}

_api_resource = {
    "Type": "AWS::ApiGateway::Resource",
    "Properties": {
        "RestApiId": {"Ref": "RestApi"},
        "ParentId": {"Fn::GetAtt": "RestApi.RootResourceId"},
        "PathPart": "test",
    },
}

_api_model = {
    "Type": "AWS::ApiGateway::Model",
    "Properties": {
        "RestApiId": {"Ref": "RestApi"},
        "ContentType": "application/json",
        "Description": "Schema for Pets example",
        "Name": "PetsModelNoFlatten",
        "Schema": {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "PetsModelNoFlatten",
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "number": {"type": "integer"},
                    "class": {"type": "string"},
                    "salesPrice": {"type": "number"},
                },
            },
        },
    },
}

_api_stage = {
    "Type": "AWS::ApiGateway::Stage",
    "Properties": {
        "StageName": "Prod",
        "Description": "Prod Stage",
        "RestApiId": {"Ref": "RestApi"},
    },
}


@pytest.mark.parametrize(
    "template,path,expected",
    [
        (
            {
                "Resources": {
                    "RestApi": _rest_api,
                }
            },
            {
                "path": deque(["Resources", "RestApi", "Properties", "Body"]),
            },
            [],
        ),
        (
            {
                "Resources": {
                    "RestApi": _rest_api,
                    "ProdStage": _api_stage,
                },
                "Outputs": {"RestApiId": {"Value": {"Ref": "RestApi"}}},
            },
            {
                "path": deque(["Resources", "RestApi", "Properties", "Body"]),
            },
            [],
        ),
        (
            {
                "Resources": {
                    "RestApi": _rest_api,
                    "RootResource": _api_resource,
                }
            },
            {
                "path": deque(["Resources", "RestApi", "Properties", "Body"]),
            },
            [
                ValidationError(
                    (
                        "Defining 'Body' with a relation to "
                        "resource 'RootResource' of type "
                        "'AWS::ApiGateway::Resource' may result "
                        "in drift and orphaned resources"
                    ),
                    rule=RestApiMixingDefinitions(),
                    path=deque([]),
                )
            ],
        ),
        (
            {
                "Resources": {
                    "RestApi": _rest_api,
                    "RootResource": _api_resource,
                }
            },
            {
                "path": deque(["Resources", "RestApi"]),
            },
            [],
        ),
        (
            {
                "Resources": {
                    "RestApi": _rest_api,
                    "RootResource": _api_resource,
                    "Model": _api_model,
                }
            },
            {
                "path": deque(["Resources", "RestApi", "Properties", "Body"]),
            },
            [
                ValidationError(
                    (
                        "Defining 'Body' with a relation to "
                        "resource 'RootResource' of type "
                        "'AWS::ApiGateway::Resource' may result "
                        "in drift and orphaned resources"
                    ),
                    rule=RestApiMixingDefinitions(),
                ),
                ValidationError(
                    (
                        "Defining 'Body' with a relation to "
                        "resource 'Model' of type "
                        "'AWS::ApiGateway::Model' may result "
                        "in drift and orphaned resources"
                    ),
                    rule=RestApiMixingDefinitions(),
                ),
            ],
        ),
    ],
    indirect=["template", "path"],
)
def test_validate(template, path, expected, rule, validator):
    errs = list(rule.validate(validator, "", "", {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class Join(BaseFn):
    """Check if Join values are correct"""

    id = "E1022"
    shortdesc = "Join validation of parameters"
    description = "Making sure the join function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-join.html"
    tags = ["functions", "join"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Join",
            ("string",),
            resolved_rule="W1032",
        )
        self.child_rules.update(
            {
                "I1022": None,
            }
        )

    def fn_join(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        validator = validator.evolve(
            context=validator.context.evolve(strict_types=True)
        )
        errs = list(super().validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)

        for rule in self.child_rules.values():
            if rule is None or not hasattr(rule, "validate"):
                continue

            yield from rule.validate(validator, s, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules._rule import CloudFormationLintRule


class ConfigError(CloudFormationLintRule):

    id = "E0003"
    shortdesc = "Error with cfn-lint configuration"
    description = "Error as a result of the cfn-lint configuration"
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["base", "rule"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError

# ruff: noqa: E501
from cfnlint.rules.resources.elasticloadbalancingv2.TargetGroupHealthCheckProtocolRestrictions import (
    TargetGroupHealthCheckProtocolRestrictions,
)


@pytest.fixture(scope="module")
def rule():
    rule = TargetGroupHealthCheckProtocolRestrictions()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"HealthCheckProtocol": "HTTPS", "Matcher": {}},
            [],
        ),
        (
            {"HealthCheckProtocol": "TCP"},
            [],
        ),
        (
            [],
            [],
        ),
        (
            {"HealthCheckProtocol": "TCP", "Matcher": {}},
            [
                ValidationError(
                    "Additional properties are not allowed ('Matcher' was unexpected)",
                    rule=TargetGroupHealthCheckProtocolRestrictions(),
                    validator="additionalProperties",
                    path=deque(["Matcher"]),
                    schema_path=deque(
                        ["then", "allOf", 0, "else", "properties", "Matcher"]
                    ),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.helpers.get_resource_by_name import get_resource_by_name
from cfnlint.rules.helpers.get_value_from_path import get_value_from_path

__all__ = [
    "get_value_from_path",
    "get_resource_by_name",
]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_ec2_instance
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class PrivateIpWithNetworkInterface(CfnLintJsonSchema):
    id = "E3674"
    shortdesc = "Primary cannoy be True when PrivateIpAddress is specified"
    description = "Only specify the private IP address for an instance in one spot"
    tags = ["resources", "ec2"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::EC2::Instance/Properties",
                "Resources/AWS::EC2::NetworkInterface/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ec2_instance,
                filename="privateipaddress.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "'Primary' cannot be True when 'PrivateIpAddress' is specified"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.resources.iam.Permissions import Permissions


@pytest.fixture(scope="module")
def rule():
    rule = Permissions()
    yield rule


@pytest.fixture
def template():
    return {
        "Transform": "AWS::Serverless-2016-10-31",
    }


@pytest.mark.parametrize(
    "name,instance,err_count",
    [
        ("Empty string", "", 0),
    ],
)
def test_permissions(name, instance, err_count, rule, validator):
    errors = list(rule.validate(validator, {}, instance, {}))
    assert len(errors) == err_count, f"Test {name!r} got {errors!r}"



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import itertools
import logging
import traceback
from functools import lru_cache
from typing import Any, Iterator, Set, Tuple

from sympy import And, Implies, Not, Symbol
from sympy.assumptions.cnf import EncodedCNF
from sympy.logic.boolalg import BooleanFalse, BooleanTrue
from sympy.logic.inference import satisfiable

from cfnlint.conditions._condition import ConditionNamed
from cfnlint.conditions._equals import Equal, EqualParameter
from cfnlint.conditions._errors import UnknownSatisfisfaction
from cfnlint.conditions._rule import Rule
from cfnlint.conditions._utils import get_hash

LOGGER = logging.getLogger(__name__)


class Conditions:
    """Conditions provides the logic for relating individual condition together"""

    _max_scenarios: int = 128  # equivalent to 2^7

    def __init__(self, cfn) -> None:
        self._conditions: dict[str, ConditionNamed] = {}
        self._parameters: dict[str, list[str]] = {}
        self._rules: list[Rule] = []
        self._init_conditions(cfn=cfn)
        self._init_parameters(cfn=cfn)
        self._init_rules(cfn=cfn)
        self._cnf, self._solver_params = self._build_cnf(list(self._conditions.keys()))

    def _init_conditions(self, cfn):
        conditions = cfn.template.get("Conditions")
        if isinstance(conditions, dict):
            for k, _ in conditions.items():
                try:
                    self._conditions[k] = ConditionNamed(k, conditions)
                except ValueError as e:
                    LOGGER.debug(
                        "Captured error while building condition %s: %s", k, str(e)
                    )
                except Exception as e:  # pylint: disable=broad-exception-caught
                    if LOGGER.getEffectiveLevel() == logging.DEBUG:
                        error_message = traceback.format_exc()
                    else:
                        error_message = str(e)
                    LOGGER.debug(
                        "Captured unknown error while building condition %s: %s",
                        k,
                        error_message,
                    )

    def _init_parameters(self, cfn: Any) -> None:
        parameters = cfn.template.get("Parameters")
        if not isinstance(parameters, dict):
            return
        for parameter_name, parameter in parameters.items():
            if not isinstance(parameter, dict):
                continue
            allowed_values = parameter.get("AllowedValues")
            if not allowed_values or not isinstance(allowed_values, list):
                continue

            param_hash = get_hash({"Ref": parameter_name})
            self._parameters[param_hash] = []
            for allowed_value in allowed_values:
                if isinstance(allowed_value, (str, int, float, bool)):
                    self._parameters[param_hash].append(get_hash(str(allowed_value)))

    def _init_rules(self, cfn: Any) -> None:
        rules = cfn.template.get("Rules")
        conditions = cfn.template.get("Conditions", {})
        if not isinstance(rules, dict) or not isinstance(conditions, dict):
            return
        for k, v in rules.items():
            if not isinstance(v, dict):
                continue
            try:
                self._rules.append(Rule(v, conditions))
            except ValueError as e:
                LOGGER.debug("Captured error while building rule %s: %s", k, str(e))
            except Exception as e:  # pylint: disable=broad-exception-caught
                if LOGGER.getEffectiveLevel() == logging.DEBUG:
                    error_message = traceback.format_exc()
                else:
                    error_message = str(e)
                LOGGER.debug(
                    "Captured unknown error while building rule %s: %s",
                    k,
                    error_message,
                )

    def get(self, name: str, default: Any = None) -> ConditionNamed:
        """Return the conditions"""
        return self._conditions.get(name, default)

    def _build_cnf(
        self, condition_names: list[str]
    ) -> Tuple[EncodedCNF, dict[str, Any]]:
        cnf = EncodedCNF()

        # build parameters and equals into solver
        equal_vars: dict[str, Symbol | BooleanFalse | BooleanTrue] = {}

        equals: dict[str, Equal] = {}

        def _build_equal_vars(c_equals: list[Equal]):
            for c_equal in c_equals:
                if c_equal.hash in equal_vars:
                    continue

                if c_equal.is_static is not None:
                    if c_equal.is_static:
                        equal_vars[c_equal.hash] = BooleanTrue()
                    else:
                        equal_vars[c_equal.hash] = BooleanFalse()
                else:
                    equal_vars[c_equal.hash] = Symbol(c_equal.hash)
                    # See if parameter in this equals is the same as another equals
                    for param in c_equal.parameters:
                        for e_hash, e_equals in equals.items():
                            if param in e_equals.parameters:
                                # equivalent to NAND logic. We want to make
                                # sure that both equals are not both True
                                # at the same time
                                cnf.add_prop(
                                    ~(equal_vars[c_equal.hash] & equal_vars[e_hash])
                                )
                equals[c_equal.hash] = c_equal

        for rule in self._rules:
            _build_equal_vars(rule.equals)

        for condition_name in condition_names:
            _build_equal_vars(self._conditions[condition_name].equals)

        # Determine if a set of conditions can never be all false
        allowed_values = self._parameters.copy()
        if allowed_values:
            # iteration 1 cleans up all the hash values
            # from allowed_values to know if we
            # used them all
            for _, equal_1 in equals.items():
                for param in equal_1.parameters:
                    if param.hash not in allowed_values:
                        continue
                    if isinstance(equal_1.left, str):
                        if get_hash(equal_1.left) in allowed_values[param.hash]:
                            allowed_values[param.hash].remove(get_hash(equal_1.left))
                        else:
                            equal_vars[equal_1.hash] = BooleanFalse()
                    if isinstance(equal_1.right, str):
                        if get_hash(equal_1.right) in allowed_values[param.hash]:
                            allowed_values[param.hash].remove(get_hash(equal_1.right))
                        else:
                            equal_vars[equal_1.hash] = BooleanFalse()

            # iteration 2 builds the cnf formulas to make sure any empty lists
            # are now full not equals
            for allowed_hash, allowed_value in allowed_values.items():
                # means the list is empty and all allowed values are validated
                # so not all equals can be false
                if not allowed_value:
                    prop = None
                    for _, equal_1 in equals.items():
                        for param in equal_1.parameters:
                            if allowed_hash == param.hash:
                                if prop is None:
                                    prop = Not(equal_vars[equal_1.hash])
                                else:
                                    prop = prop & Not(equal_vars[equal_1.hash])
                    # Need to make sure they aren't all False
                    # So Not(Not(Equal1) & Not(Equal2))
                    # When Equal1 False and Equal2 False
                    # Not(True & True) = False allowing this not to happen
                    if prop is not None:
                        cnf.add_prop(Not(prop))

        for rule in self._rules:
            cnf.add_prop(rule.build_cnf(equal_vars))

        return (cnf, equal_vars)

    def build_scenarios(
        self, conditions: dict[str, Set[bool]], region: str | None = None
    ) -> Iterator[dict[str, bool]]:
        """Given a list of condition names this function will
        yield scenarios that represent those conditions and
        there result (True/False)

        Args:
            condition_names (list[str]): A list of condition names

        Returns:
            Iterator[dict[str, bool]]: yield dict objects of {ConditionName: True/False}
        """
        # nothing to yield if there are no conditions
        if len(conditions) == 0:
            return

        c_cnf = self._cnf.copy()
        condition_names = []
        conditions_set = {}
        for condition_name, values in conditions.items():
            if condition_name in self._conditions:
                if values == {True}:
                    c_cnf.add_prop(
                        self._conditions[condition_name].build_true_cnf(
                            self._solver_params
                        )
                    )
                    conditions_set[condition_name] = True
                    continue
                if values == {False}:
                    c_cnf.add_prop(
                        self._conditions[condition_name].build_false_cnf(
                            self._solver_params
                        )
                    )
                    conditions_set[condition_name] = False
                    continue
            condition_names.append(condition_name)

        try:
            # build a large matric of True/False options
            # based on the provided conditions
            scenarios_attempted = 0
            if region:
                products = itertools.starmap(
                    self.build_scenerios_on_region,
                    itertools.product(condition_names, [region]),
                )
            else:
                products = itertools.product([True, False], repeat=len(condition_names))  # type: ignore

            for p in products:
                cnf = c_cnf.copy()
                params = dict(zip(condition_names, p))
                for condition_name, opt in params.items():
                    if opt:
                        cnf.add_prop(
                            self._conditions[condition_name].build_true_cnf(
                                self._solver_params
                            )
                        )
                    else:
                        cnf.add_prop(
                            self._conditions[condition_name].build_false_cnf(
                                self._solver_params
                            )
                        )

                # if the scenario can be satisfied then return it
                if satisfiable(cnf):
                    yield {**params, **conditions_set}

                scenarios_attempted += 1
                # On occassions people will use a lot of non-related conditions
                # this is fail safe to limit the maximum number of responses
                if scenarios_attempted >= self._max_scenarios:
                    return
        except KeyError:
            # KeyError is because the listed condition doesn't exist because of bad
            #  formatting or just the wrong condition name
            return

    def _build_cfn_implies(self, scenarios) -> And:
        conditions = []
        for condition_name, opt in scenarios.items():
            if opt:
                conditions.append(
                    self._conditions[condition_name].build_true_cnf(self._solver_params)
                )
            else:
                conditions.append(
                    self._conditions[condition_name].build_false_cnf(
                        self._solver_params
                    )
                )

        return And(*conditions)

    def check_implies(self, scenarios: dict[str, bool], implies: str) -> bool:
        """Based on a bunch of scenario conditions and their Truth/False value
        determine if implies condition is True any time the scenarios are satisfied
        solver, solver_params = self._build_solver(list(scenarios.keys()) + [implies])

        Args:
            scenarios (dict[str, bool]): A list of condition names
            and if they are True or False implies: the condition name that
            we are implying will also be True

        Returns:
            bool: if the implied condition will be True if the scenario is True
        """
        try:
            cnf = self._cnf.copy()
            # if the implies condition has to be false in the scenarios we
            # know it can never be true
            if not scenarios.get(implies, True):
                return False

            and_condition = self._build_cfn_implies(scenarios)
            cnf.add_prop(and_condition)
            implies_condition = self._conditions[implies].build_true_cnf(
                self._solver_params
            )
            cnf.add_prop(Not(Implies(and_condition, implies_condition)))

            results = satisfiable(cnf)
            if results:
                return False

            return True
        except KeyError:
            # KeyError is because the listed condition doesn't exist because of bad
            #  formatting or just the wrong condition name
            return True

    @lru_cache
    def build_scenerios_on_region(self, condition_name: str, region: str) -> list[bool]:
        """Based on a region validate if the condition_name could be true

        Args:
            condition_name (str): The name of the condition we are validating against
            region (str): the name of the region

        Returns:
            list[bool]: Returns True, False, or True and False depending on if the
               condition could be True, False or both based on the region parameter
        """
        cnf_region = self._cnf.copy()
        found_region = False

        # validate the condition name exists
        # return True/False if it doesn't
        if condition_name not in self._conditions:
            return [True, False]

        for eql in self._conditions[condition_name].equals:
            is_region, equal_region = eql.is_region
            if is_region:
                found_region = True
                if equal_region == region:
                    cnf_region.add_prop(And(self._solver_params[eql.hash]))
                else:
                    cnf_region.add_prop(Not(self._solver_params[eql.hash]))

        # The condition doesn't use a region parameter so it can be True or False
        # Note: It is possible its a hard coded condition but
        # for now we will return True and False
        if not found_region:
            return [True, False]

        results = []
        cnf_test = cnf_region.copy()
        cnf_test.add_prop(
            self._conditions[condition_name].build_true_cnf(self._solver_params)
        )
        if satisfiable(cnf_test):
            results.append(True)

        cnf_test = cnf_region.copy()
        cnf_test.add_prop(
            self._conditions[condition_name].build_false_cnf(self._solver_params)
        )
        if satisfiable(cnf_test):
            results.append(False)

        return results

    def satisfiable(
        self, conditions: dict[str, bool], parameter_values: dict[str, str]
    ) -> bool:
        """Given a list of condition names this function will
        determine if the conditions are satisfied

        Args:
            condition_names (dict[str, bool]): A list of condition names with if
              they are True or False

        Returns:
            bool: True if the conditions are satisfied

        Raises:
            UnknownSatisfisfaction: If we don't know how to satisfy a condition
        """
        if not conditions:
            if self._rules:
                satisfied = satisfiable(self._cnf, all_models=False)
                if satisfied is False:
                    return satisfied
                return True
            else:
                return True

        cnf = self._cnf.copy()
        at_least_one_param_found = False

        for condition_name, opt in conditions.items():
            for c_equals in self._conditions[condition_name].equals:
                found_params = {}
                for param, value in parameter_values.items():
                    ref_hash = get_hash({"Ref": param})

                    for c_equal_param in c_equals.parameters:
                        if isinstance(c_equal_param, EqualParameter):
                            if c_equal_param.satisfiable is False:
                                raise UnknownSatisfisfaction(
                                    f"Can't resolve satisfaction for {condition_name!r}"
                                )

                    if ref_hash in c_equals.parameters:
                        found_params = {ref_hash: value}

                if not found_params:
                    continue

                at_least_one_param_found = True
                if c_equals.test(found_params):
                    cnf.add_prop(Symbol(c_equals.hash))
                else:
                    cnf.add_prop(Not(Symbol(c_equals.hash)))

                if opt:
                    cnf.add_prop(
                        self._conditions[condition_name].build_true_cnf(
                            self._solver_params
                        )
                    )
                else:
                    cnf.add_prop(
                        self._conditions[condition_name].build_false_cnf(
                            self._solver_params
                        )
                    )

        if at_least_one_param_found is False:
            if self._rules:
                satisfied = satisfiable(self._cnf, all_models=False)
                if satisfied is False:
                    return satisfied
                return True
            else:
                return True

        satisfied = satisfiable(cnf, all_models=False)
        if satisfied is False:
            return satisfied
        return True



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Any, Mapping, Sequence
from urllib.parse import unquote, urldefrag, urljoin

from cfnlint.schema.resolver._exceptions import RefResolutionError
from cfnlint.schema.resolver._utils import id_of

_SUBSCHEMAS_KEYWORDS = ("$id", "id", "$anchor", "$dynamicAnchor")


def _match_subschema_keywords(value):
    for keyword in _SUBSCHEMAS_KEYWORDS:
        if keyword in value:
            yield keyword, value


def _search_schema(schema, matcher):
    """Breadth-first search routine."""
    values = deque([schema])
    while values:
        value = values.pop()
        if not isinstance(value, dict):
            continue
        yield from matcher(value)
        values.extendleft(value.values())


def _match_keyword(keyword):
    def matcher(value):
        if keyword in value:
            yield value

    return matcher


@dataclass
class RefResolver:
    referrer: Any = field(init=True)
    base_uri: str = field(init=True, default="")
    _id_of: Any = field(init=False, default=id_of)
    _scopes_stack: Any = field(init=False)
    _urljoin_cache: Any = field(init=True, default=None)
    _cache: Any = field(init=True, default=None)
    _cache_cfn_pointer: Any = field(init=True, default=None)
    store: Any = field(init=True, default=None)

    def __post_init__(self):
        self._scopes_stack = [self.base_uri]
        if self._cache is None:
            self._cache = lru_cache(1024)(self.resolve_from_url)
        if self._cache_cfn_pointer is None:
            self._cache_cfn_pointer = lru_cache(1024)(self.resolve_cfn_pointer)
        if self._urljoin_cache is None:
            self._urljoin_cache = lru_cache(1024)(urljoin)
        if self.store is None:
            self.store = {}
            self.store.update(
                (schema["$id"], schema)
                for schema in self.store.values()
                if isinstance(schema, Mapping) and "$id" in schema
            )

    @classmethod
    def from_schema(cls, schema, **kwargs) -> "RefResolver":
        """
        Construct a resolver from a JSON schema object.

        Arguments:

            schema:

                the referring schema

        Returns:

            `RefResolver`
        """
        return cls(base_uri=id_of(schema), referrer=schema, **kwargs)

    def push_scope(self, scope):
        """
        Enter a given sub-scope.

        Treats further dereferences as being performed underneath the
        given scope.
        """
        self._scopes_stack.append(
            self._urljoin_cache(self.resolution_scope, scope),
        )

    def pop_scope(self):
        """
        Exit the most recent entered scope.

        Treats further dereferences as being performed underneath the
        original scope.

        Don't call this method more times than `push_scope` has been
        called.
        """
        try:
            self._scopes_stack.pop()
        except IndexError as e:
            raise RefResolutionError(
                "Failed to pop the scope from an empty stack. "
                "`pop_scope()` should only be called once for every "
                "`push_scope()`",
            ) from e

    def resolve_from_url(self, url):
        """
        Resolve the given URL.
        """
        url, fragment = urldefrag(url)
        if url:
            try:
                document = self.store[url]
            except KeyError as e:
                raise RefResolutionError(
                    "Failed to pop the scope from an empty stack. "
                    "`pop_scope()` should only be called once for every "
                    "`push_scope()`",
                ) from e
        else:
            document = self.referrer

        return self.resolve_fragment(document, fragment)

    def resolve(self, ref):
        """
        Resolve the given reference.
        """
        url = self._urljoin_cache(self.resolution_scope, ref).rstrip("/")
        match = self._find_in_subschemas(url)
        if match is not None:
            return match

        return url, self._cache(url)

    def resolve_cfn_pointer(self, pointer: str) -> dict[str, Any]:
        """
        Resolve the given cfn pointer.

        /properties/Tags/*/Name
        /properties/SourceConfiguration/ApplicationName
        """

        # cfn pointers don't cross schemas so there is no remote URL to
        # to worry about

        schema = self.referrer

        # to handle custom resources
        if pointer == "/properties/CfnLintAllTypes":
            return {
                "type": [
                    "string",
                    "integer",
                    "number",
                    "boolean",
                    "object",
                    "array",
                ]
            }
        elif pointer == "/properties/CfnLintStringType":
            return {
                "type": [
                    "string",
                ]
            }

        ref = pointer.lstrip("/").split("/") if pointer else []

        return self._walk_cfn_pointer(schema, ref[1:])

    @property
    def resolution_scope(self):
        """
        Retrieve the current resolution scope.
        """
        return self._scopes_stack[-1]

    def _walk_cfn_pointer(
        self, document: dict[str, Any], pointer: Sequence[str]
    ) -> dict[str, Any]:
        """
        Resolve a cfn ``pointer`` within the referenced ``document``.

        Arguments:

            document:

                The referent document

            pointer (str):

                a CFN pointer to resolve within it
        """

        # depth first takes precedence
        if isinstance(document, dict) and "$ref" in document:
            _, document = self.resolve(document.get("$ref"))
            return self._walk_cfn_pointer(document, pointer)

        if not pointer:
            return document

        point = pointer[0]

        try:
            if point == "*":
                if "items" in document:
                    return self._walk_cfn_pointer(document["items"], pointer[1:])
            return self._walk_cfn_pointer(document["properties"][point], pointer[1:])
        except (TypeError, LookupError, KeyError, RefResolutionError) as e:
            for c in ["anyOf", "allOf", "oneOf"]:
                if c in document:
                    for i in document[c]:
                        try:
                            return self._walk_cfn_pointer(i, pointer)
                        except RefResolutionError:
                            continue

            if e.__class__.__name__ == "RefResolutionError":
                raise e

            raise RefResolutionError(f"Unresolvable CFN pointer: {point!r}") from e

    def resolve_fragment(self, document, fragment):
        """
        Resolve a ``fragment`` within the referenced ``document``.

        Arguments:

            document:

                The referent document

            fragment (str):

                a URI fragment to resolve within it
        """
        fragment = fragment.lstrip("/")

        if not fragment:
            return document

        def find(key):
            yield from _search_schema(document, _match_keyword(key))

        for keyword in ["id", "$id"]:
            for subschema in find(keyword):
                if "#" + fragment == subschema[keyword]:
                    return subschema

        # Resolve via path
        parts = unquote(fragment).split("/") if fragment else []
        for part in parts:
            part = part.replace("~1", "/").replace("~0", "~")

            if isinstance(document, Sequence):
                # Array indexes should be turned into integers
                try:
                    part = int(part)
                except ValueError:
                    pass
            try:
                document = document[part]
            except (TypeError, LookupError) as e:
                raise RefResolutionError(
                    f"Unresolvable JSON pointer: {fragment!r}"
                ) from e

        return document

    def _get_subschemas_cache(self):
        cache = {key: [] for key in _SUBSCHEMAS_KEYWORDS}
        for keyword, subschema in _search_schema(
            self.referrer,
            _match_subschema_keywords,
        ):
            cache[keyword].append(subschema)
        return cache

    def _find_in_subschemas(self, url):
        uri, fragment = urldefrag(url)
        subschemas = self._get_subschemas_cache()["$id"]
        if not subschemas:
            return None

        for subschema in subschemas:
            target_uri = self._urljoin_cache(
                self.resolution_scope,
                subschema["$id"],
            )
            if target_uri.rstrip("/") == uri.rstrip("/"):
                if fragment:
                    subschema = self.resolve_fragment(subschema, fragment)
                return url, subschema
        return None



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbClusterAuroraWarning import DbClusterAuroraWarning


@pytest.fixture(scope="module")
def rule():
    rule = DbClusterAuroraWarning()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"Engine": "aurora-mysql", "EngineMode": "serverless"},
            [],
        ),
        (
            {
                "Engine": "aurora-mysql",
                "EngineMode": "provisioned",
                "PerformanceInsightsEnabled": True,
            },
            [],
        ),
        (
            {
                "Engine": "aurora-mysql",
                "PerformanceInsightsEnabled": True,
            },
            [],
        ),
        (
            {
                "Engine": "aurora-mysql",
                "EngineMode": "serverless",
                "PerformanceInsightsEnabled": True,
            },
            [
                ValidationError(
                    (
                        "Additional properties are not allowed "
                        "'PerformanceInsightsEnabled' when creating Aurora cluster"
                    ),
                    rule=DbClusterAuroraWarning(),
                    path=deque(["PerformanceInsightsEnabled"]),
                    validator=None,
                    schema_path=deque(
                        ["then", "properties", "PerformanceInsightsEnabled"]
                    ),
                ),
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Helpers for loading resources, managing specs, constants, etc.

Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import datetime
import fnmatch
import gzip
import hashlib
import importlib
import importlib.resources as pkg_resources
import inspect
import json
import logging
import os
from io import BytesIO
from typing import Any, Sequence
from urllib.request import Request, urlopen, urlretrieve

import regex as re

LOGGER = logging.getLogger(__name__)

AVAILABILITY_ZONES = {
    "af-south-1": ["af-south-1a", "af-south-1b", "af-south-1c"],
    "ap-east-1": ["ap-east-1a", "ap-east-1b", "ap-east-1c"],
    "ap-northeast-1": [
        "ap-northeast-1a",
        "ap-northeast-1b",
        "ap-northeast-1c",
        "ap-northeast-1d",
    ],
    "ap-northeast-2": [
        "ap-northeast-2a",
        "ap-northeast-2b",
        "ap-northeast-2c",
        "ap-northeast-2d",
    ],
    "ap-northeast-3": ["ap-northeast-3a", "ap-northeast-3b", "ap-northeast-3c"],
    "ap-south-1": ["ap-south-1a", "ap-south-1b", "ap-south-1c"],
    "ap-south-2": [
        "ap-south-2a",
        "ap-south-2b",
        "ap-south-2c",
    ],
    "ap-southeast-1": ["ap-southeast-1a", "ap-southeast-1b", "ap-southeast-1c"],
    "ap-southeast-2": ["ap-southeast-2a", "ap-southeast-2b", "ap-southeast-2c"],
    "ap-southeast-3": ["ap-southeast-3a", "ap-southeast-3b", "ap-southeast-3c"],
    "ap-southeast-4": [
        "ap-southeast-4a",
        "ap-southeast-4b",
        "ap-southeast-4c",
    ],
    "ap-southeast-5": [
        "ap-southeast-5a",
        "ap-southeast-5b",
        "ap-southeast-5c",
    ],
    "ap-southeast-7": [
        "ap-southeast-7a",
        "ap-southeast-7b",
        "ap-southeast-7c",
    ],
    "ca-west-1": [
        "ca-west-1a",
        "ca-west-1b",
        "ca-west-1c",
    ],
    "ca-central-1": [
        "ca-central-1a",
        "ca-central-1b",
        "ca-central-1c",
        "ca-central-1d",
    ],
    "cn-north-1": ["cn-north-1a", "cn-north-1b", "cn-north-1c"],
    "cn-northwest-1": ["cn-northwest-1a", "cn-northwest-1b", "cn-northwest-1c"],
    "eu-central-1": ["eu-central-1a", "eu-central-1b", "eu-central-1c"],
    "eu-central-2": [
        "eu-central-2a",
        "eu-central-2b",
        "eu-central-2c",
    ],
    "eu-north-1": ["eu-north-1a", "eu-north-1b", "eu-north-1c"],
    "eu-south-1": ["eu-south-1a", "eu-south-1b", "eu-south-1c"],
    "eu-south-2": [
        "eu-south-2a",
        "eu-south-2b",
        "eu-south-2c",
    ],
    "eu-west-1": ["eu-west-1a", "eu-west-1b", "eu-west-1c"],
    "eu-west-2": ["eu-west-2a", "eu-west-2b", "eu-west-2c"],
    "eu-west-3": ["eu-west-3a", "eu-west-3b", "eu-west-3c"],
    "il-central-1": ["il-central-1a", "il-central-1b", "il-central-1c"],
    "me-south-1": ["me-south-1a", "me-south-1b", "me-south-1c"],
    "me-central-1": [
        "me-central-1a",
        "me-central-1b",
        "me-central-1c",
    ],
    "sa-east-1": ["sa-east-1a", "sa-east-1b", "sa-east-1c"],
    "us-east-1": [
        "us-east-1a",
        "us-east-1b",
        "us-east-1c",
        "us-east-1d",
        "us-east-1e",
        "us-east-1f",
    ],
    "us-east-2": ["us-east-2a", "us-east-2b", "us-east-2c"],
    "us-gov-east-1": ["us-gov-east-1a", "us-gov-east-1b", "us-gov-east-1c"],
    "us-gov-west-1": ["us-gov-west-1a", "us-gov-west-1b", "us-gov-west-1c"],
    "us-west-1": ["us-west-1a", "us-west-1c"],
    "us-west-2": ["us-west-2a", "us-west-2b", "us-west-2c", "us-west-2d"],
    # Isolated regions
    "us-iso-east-1": ["us-iso-east-1a", "us-iso-east-1b", "us-iso-east-1c"],
    "us-iso-west-1": ["us-iso-west-1a", "us-iso-west-1b", "us-iso-west-1c"],
    "us-isob-east-1": ["us-isob-east-1a", "us-isob-east-1b", "us-isob-east-1c"],
    "us-isof-east-1": ["us-isof-east-1a", "us-isof-east-1b", "us-isof-east-1c"],
    "us-isof-south-1": ["us-isof-south-1a", "us-isof-south-1b", "us-isof-south-1c"],
    "eu-isoe-west-1": ["eu-isoe-west-1a", "eu-isoe-west-1b", "eu-isoe-west-1c"],
}

REGIONS = list(AVAILABILITY_ZONES.keys())
REGION_PRIMARY = "us-east-1"
TAG_MAP = "tag:yaml.org,2002:map"
UNCONVERTED_SUFFIXES = ["Ref", "Condition"]
FN_PREFIX = "Fn::"
CONDITION_FUNCTIONS = ["Fn::If"]

REGEX_ALPHANUMERIC = re.compile("^[a-zA-Z0-9]*$")
REGEX_CIDR = re.compile(
    r"^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(\/([0-9]|[1-2][0-9]|3[0-2]))$"
)
REGEX_IPV4 = re.compile(
    r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)){3}$"
)
REGEX_IPV6 = re.compile(
    r"^(((?=.*(::))(?!.*\3.+\3))\3?|[\dA-F]{1,4}:)([\dA-F]{1,4}(\3|:\b)|\2){5}(([\dA-F]{1,4}(\3|:\b|$)|\2){2}|(((2[0-4]|1\d|[1-9])?\d|25[0-5])\.?\b){4})\Z",
    re.I | re.S,
)
REGEX_DYN_REF = re.compile(r"^.*{{\s*(resolve:.+)\s*}}.*$")
REGEX_DYN_REF_SSM = re.compile(r"^.*{{resolve:ssm:[a-zA-Z0-9_\.\-/]+(:\d+)?}}.*$")
REGEX_DYN_REF_SSM_SECURE = re.compile(
    r"^.*{{resolve:ssm-secure:[a-zA-Z0-9_\.\-/]+(:\d+)?}}.*$"
)
REGEX_SUB_PARAMETERS = re.compile(r"\${\s*([^!\s].*?)\s*}")

FUNCTIONS = frozenset(
    [
        "Fn::Base64",
        "Fn::Cidr",
        "Fn::Contains",
        "Fn::FindInMap",
        "Fn::ForEach::[a-zA-Z0-9]+",
        "Fn::GetAtt",
        "Fn::GetAZs",
        "Fn::If",
        "Fn::ImportValue",
        "Fn::Join",
        "Fn::Length",
        "Fn::Select",
        "Fn::Split",
        "Fn::Sub",
        "Fn::ToJsonString",
        "Fn::Transform",
        "Ref",
    ]
)

FUNCTIONS_LIST = frozenset(
    [
        "Fn::GetAZs",
        "Fn::Split",
        "Fn::Cidr",
        "Fn::GetAtt",
        "Fn::FindInMap",
        "Fn::Select",
        "Ref",
    ]
)
FUNCTIONS_OBJECT = frozenset(["Fn::Select", "Fn::GetAtt"])
# FindInMap can be singular or multiple.  This needs to be accounted for individually
# GetAtt can refere to a list, object, or a singular value
FUNCTIONS_SINGLE = frozenset(
    set(FUNCTIONS)
    - (set(FUNCTIONS_LIST) - set(["Fn::FindInMap", "Fn::GetAtt", "Fn::Select", "Ref"]))
)

FUNCTION_IF = "Fn::If"
FUNCTION_AND = "Fn::And"
FUNCTION_OR = "Fn::Or"
FUNCTION_NOT = "Fn::Not"
FUNCTION_EQUALS = "Fn::Equals"
FUNCTION_BASE64 = "Fn::Base64"
FUNCTION_CONTAINS = "Fn::Contains"
FUNCTION_EACH_MEMBER_EQUALS = "Fn::EachMemberEquals"
FUNCTION_EACH_MEMBER_IN = "Fn::EachMemberIn"
FUNCTION_REF_ALL = "Fn::RefAll"
FUNCTION_VALUE_OF = "Fn::ValueOf"
FUNCTION_VALUE_OF_ALL = "Fn::ValueOfAll"
FUNCTION_FOR_EACH = re.compile(r"^Fn::ForEach::[a-zA-Z0-9]+$")
FUNCTION_TRANSFORM = "Fn::Transform"

FUNCTION_CONDITIONS = frozenset(
    [FUNCTION_AND, FUNCTION_OR, FUNCTION_NOT, FUNCTION_EQUALS]
)

FUNCTION_RULES = frozenset(
    [
        FUNCTION_AND,
        FUNCTION_OR,
        FUNCTION_NOT,
        FUNCTION_EQUALS,
        FUNCTION_CONTAINS,
        FUNCTION_EACH_MEMBER_EQUALS,
        FUNCTION_EACH_MEMBER_IN,
        FUNCTION_REF_ALL,
        FUNCTION_VALUE_OF,
        FUNCTION_VALUE_OF_ALL,
    ]
)

FUNCTIONS_ALL = frozenset.union(
    *[FUNCTIONS, FUNCTION_CONDITIONS, frozenset(["Condition"])]
)

PSEUDOPARAMS_SINGLE = frozenset(
    [
        "AWS::AccountId",
        "AWS::Partition",
        "AWS::Region",
        "AWS::StackId",
        "AWS::StackName",
        "AWS::URLSuffix",
    ]
)

PSEUDOPARAMS_MULTIPLE = frozenset(
    [
        "AWS::NotificationARNs",
    ]
)

PSEUDOPARAMS = frozenset(
    ["AWS::NoValue"] + list(PSEUDOPARAMS_SINGLE) + list(PSEUDOPARAMS_MULTIPLE)
)

LIMITS: dict[str, Any] = {
    "Mappings": {"number": 200, "attributes": 200, "name": 255},  # in characters
    "Outputs": {
        "number": 200,
        "name": 255,  # in characters
        "description": 1024,  # in bytes
    },
    "Parameters": {
        "number": 200,
        "name": 255,  # in characters
        "value": 4096,  # in bytes
    },
    "Resources": {"number": 500, "name": 255},  # in characters
    "template": {"body": 1000000},  # in bytes  # in bytes
    "threshold": 0.9,  # for rules about approaching the other limit values
}

valid_snapshot_types = frozenset(
    [
        "AWS::DocDB::DBCluster",
        "AWS::EC2::Volume",
        "AWS::ElastiCache::CacheCluster",
        "AWS::ElastiCache::ReplicationGroup",
        "AWS::Neptune::DBCluster",
        "AWS::RDS::DBCluster",
        "AWS::RDS::DBInstance",
        "AWS::Redshift::Cluster",
    ]
)

VALID_PARAMETER_TYPES_SINGLE = frozenset(
    [
        "AWS::EC2::AvailabilityZone::Name",
        "AWS::EC2::Image::Id",
        "AWS::EC2::Instance::Id",
        "AWS::EC2::KeyPair::KeyName",
        "AWS::EC2::SecurityGroup::GroupName",
        "AWS::EC2::SecurityGroup::Id",
        "AWS::EC2::Subnet::Id",
        "AWS::EC2::VPC::Id",
        "AWS::EC2::Volume::Id",
        "AWS::Route53::HostedZone::Id",
        "AWS::SSM::Parameter::Name",
        "Number",
        "String",
        "AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>",
        "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>",
        "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>",
        "AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>",
        "AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>",
        "AWS::SSM::Parameter::Value<AWS::SSM::Parameter::Name>",
        "AWS::SSM::Parameter::Value<Number>",
        "AWS::SSM::Parameter::Value<String>",
    ]
)

VALID_PARAMETER_TYPES_LIST = frozenset(
    [
        "CommaDelimitedList",
        "List<AWS::EC2::AvailabilityZone::Name>",
        "List<AWS::EC2::Image::Id>",
        "List<AWS::EC2::Instance::Id>",
        "List<AWS::EC2::SecurityGroup::GroupName>",
        "List<AWS::EC2::SecurityGroup::Id>",
        "List<AWS::EC2::Subnet::Id>",
        "List<AWS::EC2::VPC::Id>",
        "List<AWS::EC2::Volume::Id>",
        "List<AWS::Route53::HostedZone::Id>",
        "List<Number>",
        "List<String>",
        "AWS::SSM::Parameter::Value<CommaDelimitedList>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>",
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>",
        "AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>",
        "AWS::SSM::Parameter::Value<List<Number>>",
        "AWS::SSM::Parameter::Value<List<String>>",
    ]
)

TEMPLATED_PROPERTY_CFN_PATHS = frozenset(
    [
        "Resources/AWS::ApiGateway::RestApi/Properties/BodyS3Location",
        "Resources/AWS::Lambda::Function/Properties/Code",
        "Resources/AWS::Lambda::LayerVersion/Properties/Content",
        "Resources/AWS::ElasticBeanstalk::ApplicationVersion/Properties/SourceBundle",
        "Resources/AWS::StepFunctions::StateMachine/Properties/DefinitionS3Location",
        "Resources/AWS::AppSync::GraphQLSchema/Properties/DefinitionS3Location",
        "Resources/AWS::AppSync::Resolver/Properties/RequestMappingTemplateS3Location",
        "Resources/AWS::AppSync::Resolver/Properties/ResponseMappingTemplateS3Location",
        "Resources/AWS::AppSync::FunctionConfiguration/Properties/RequestMappingTemplateS3Location",
        "Resources/AWS::AppSync::FunctionConfiguration/Properties/ResponseMappingTemplateS3Location",
        "Resources/AWS::CloudFormation::Stack/Properties/TemplateURL",
        "Resources/AWS::CodeCommit::Repository/Properties/Code/S3",
    ]
)

VALID_PARAMETER_TYPES = list(VALID_PARAMETER_TYPES_SINGLE) + list(
    VALID_PARAMETER_TYPES_LIST
)

BOOLEAN_STRINGS_TRUE = frozenset(["true", "True", "TRUE"])
BOOLEAN_STRINGS_FALSE = frozenset(["false", "False", "FALSE"])
BOOLEAN_STRINGS = frozenset(list(BOOLEAN_STRINGS_TRUE) + list(BOOLEAN_STRINGS_FALSE))

TRANSFORM_SAM = "AWS::Serverless-2016-10-31"
TRANSFORM_LANGUAGE_EXTENSION = "AWS::LanguageExtensions"


# pylint: disable=missing-class-docstring
class RegexDict(dict):
    def __getitem__(self, item):
        possible_items = {}
        for k, v in self.items():
            if re.fullmatch(k, item):
                possible_items[k] = v
        if not possible_items:
            raise KeyError
        longest_match = sorted(possible_items.keys(), key=len)[-1]
        return possible_items[longest_match]

    def __contains__(self, item):
        if isinstance(item, (dict, list)):
            return False
        for k, _ in self.items():
            if re.fullmatch(k, item):
                return True
        return False

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default


def get_metadata_filename(url):
    """Returns the filename for a metadata file associated with a remote resource"""
    caching_dir = os.path.join(os.path.dirname(__file__), "data", "DownloadsMetadata")
    encoded_url = hashlib.sha256(url.encode()).hexdigest()
    metadata_filename = os.path.join(caching_dir, encoded_url + ".meta.json")

    return metadata_filename


def url_has_newer_version(url):
    """Checks to see if a newer version of the resource at the URL is available
    Always returns true if using Python2.7 due to lack of HEAD request support,
    or if we have no caching information for the local version of the resource
    """
    metadata_filename = get_metadata_filename(url)

    # Load in the cache
    metadata = load_metadata(metadata_filename)

    # Etag is a caching identifier used by S3 and Cloudfront
    if "etag" in metadata:
        cached_etag = metadata["etag"]
    else:
        # If we don't know the etag of the local version, we should force an update
        return True

    # Need to wrap this in a try, as URLLib2 in Python2 doesn't support HEAD requests
    try:
        # Make an initial HEAD request
        req = Request(url, method="HEAD")
        with urlopen(req) as res:
            # If we have an ETag value stored and it matches the returned one,
            # then we already have a copy of the most recent version of the
            # resource, so don't bother fetching it again
            if (
                cached_etag
                and res.info().get("ETag")
                and cached_etag == res.info().get("ETag")
            ):
                LOGGER.debug(
                    "We already have a cached version of url %s with ETag value of %s",
                    url,
                    cached_etag,
                )
                return False

    except NameError:
        # We should force an update
        return True

    # The ETag value of the remote resource does not
    # match the local one, so a newer version is available
    return True


def get_url_content(url, caching=False):
    """Get the contents of a spec file"""

    with urlopen(url) as res:
        if caching and res.info().get("ETag"):
            metadata_filename = get_metadata_filename(url)
            # Load in all existing values
            metadata = load_metadata(metadata_filename)
            metadata["etag"] = res.info().get("ETag")
            metadata["url"] = url  # To make it obvious which url the Tag relates to
            save_metadata(metadata, metadata_filename)

        # Continue to handle the file download normally
        if res.info().get("Content-Encoding") == "gzip":
            buf = BytesIO(res.read())
            f = gzip.GzipFile(fileobj=buf)
            content = f.read().decode("utf-8")
        else:
            content = res.read().decode("utf-8")

    return content


def get_url_retrieve(url: str, caching: bool = False) -> str:
    """Get the contents of a zip file and returns
    a string representing the file

    Args:
        url (str): The url to retrieve
        caching (bool): If we can cache the results (default: False)
    Returns:
        str: A string representing the file object that was retrieved
    """

    if caching:
        req = Request(url, method="HEAD")
        with urlopen(req) as res:
            if res.info().get("ETag"):
                metadata_filename = get_metadata_filename(url)
                # Load in all existing values
                metadata = load_metadata(metadata_filename)
                metadata["etag"] = res.info().get("ETag")
                metadata["url"] = url  # To make it obvious which url the Tag relates to
                save_metadata(metadata, metadata_filename)

    fileobject, _ = urlretrieve(url)

    return fileobject


def load_metadata(filename):
    """Get the contents of the download metadata file"""
    metadata = {}
    if os.path.exists(filename):
        with open(filename, "r", encoding="utf-8") as metadata_file:
            metadata = json.load(metadata_file)
    return metadata


def save_metadata(metadata, filename):
    """Save the contents of the download metadata file"""
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.mkdir(dirname)

    with open(filename, "w", encoding="utf-8") as metadata_file:
        json.dump(metadata, metadata_file)
        metadata_file.write("\n")


def load_resource(package, filename="us-east-1.json"):
    """Load CloudSpec resources
    :param filename: filename to load
    :return: Json output of the resource laoded
    """
    return json.loads(
        pkg_resources.files(package)  # pylint: disable=no-member
        .joinpath(filename)
        .read_text(encoding="utf-8")
    )


def is_custom_resource(resource_type):
    """Return True if resource_type is a custom resource"""
    return resource_type and (
        resource_type == "AWS::CloudFormation::CustomResource"
        or resource_type.startswith("Custom::")
    )


def bool_compare(first, second):
    """Compare strings to boolean values"""

    if isinstance(first, str):
        first = bool(first.lower() in ["true", "True"])

    if isinstance(second, str):
        second = bool(second.lower() in ["true", "True"])

    return first is second


def is_function(instance: Any) -> tuple[str | None, Any]:
    """
    Checks if the given instance is a dictionary representing a function.

    Args:
        instance (Any): The object to check if it represents a function.

    Returns:
        tuple[str, Any] | None: If the instance is a dictionary with a single
        key-value pair, and the key is a valid function name, returns a tuple
        containing the function name (str) and the function arguments (Any).
        Otherwise, returns a tuple with None and None.
    """
    if isinstance(instance, dict):
        if len(instance) == 1:
            for key, value in instance.items():
                if key in FUNCTIONS_ALL:
                    return key, value

    return None, None


def _translate_types(types: Sequence[str]) -> list[str]:
    """
    Return compatible types. This is an adventitious result
    meaning a string could be an integer.

    Args:
        types (Sequence[str]): The types

    Returns:
        bool: If any type of source is compatible with any type in the destination
    """
    compatible_types = []
    for t in types:
        if t == "string":
            compatible_types.extend([t, "number", "boolean", "integer"])
        if t == "integer":
            compatible_types.extend([t, "number", "string"])
        if t == "boolean":
            compatible_types.extend([t, "string"])
        if t == "number":
            compatible_types.extend([t, "string"])
        else:
            compatible_types.append(t)
    return compatible_types


def is_types_compatible(
    source_types: str | Sequence[str],
    destination_types: str | Sequence[str],
    strict_types: bool = False,
) -> bool:
    """
    Validate if desination types are compatible with source types.

    Args:
        source_types (str | Sequence[str]): The source types
        destination_types (str | Sequence[str]): The destination types

    Returns:
        bool: If any type of source is compatible with any type in the destination
    """
    if not strict_types:
        source_types = _translate_types(ensure_list(source_types))
    destination_types = ensure_list(destination_types)

    if any(schema_type in source_types for schema_type in destination_types):
        return True

    return False


def format_json_string(json_string):
    """Format the given JSON string"""

    def converter(o):  # pylint: disable=R1710
        """Help convert date/time into strings"""
        if isinstance(o, datetime.datetime):
            return o.__str__()  # pylint: disable=unnecessary-dunder-call

    return json.dumps(
        json_string, indent=1, sort_keys=True, separators=(",", ": "), default=converter
    )


def create_rules(
    mod,
    name="CloudFormationLintRule",
    modules=("cfnlint", "cfnlint.rules", "cfnlint.rules._rule"),
):
    """Create and return an instance of each CloudFormationLintRule subclass
    from the given module."""
    result = []
    for _, clazz in inspect.getmembers(mod, inspect.isclass):
        if (
            clazz.__name__ == "CustomRule"
            and clazz.__module__ == "cfnlint.rules.custom"
        ):
            continue
        method_resolution = inspect.getmro(clazz)
        if [
            clz
            for clz in method_resolution[1:]
            if clz.__module__ in modules and clz.__name__ == name
        ]:
            # create and instance of subclasses of CloudFormationLintRule
            obj = clazz()
            result.append(obj)
    return result


loader_details = (
    importlib.machinery.SourceFileLoader,  # pylint: disable=no-member
    importlib.machinery.SOURCE_SUFFIXES,  # pylint: disable=no-member
)


def import_filename(pluginname, root):
    """import_filename imports a module from a file"""
    mod_finder = importlib.machinery.FileFinder(  # pylint: disable=no-member
        root, loader_details
    )

    mod_spec = mod_finder.find_spec(pluginname)
    if mod_spec is not None:
        # for python 2.7 disabling pylint checks
        mod = importlib.util.module_from_spec(mod_spec)  # pylint: disable=no-member
        mod_spec.loader.exec_module(mod)
        return mod

    return None


def load_plugins(
    directory,
    name="CloudFormationLintRule",
    modules=("cfnlint", "cfnlint.rules", "cfnlint.rules._rule"),
):
    """Load plugins"""
    result = []

    def onerror(os_error):
        """Raise an error"""
        raise os_error

    for root, _, filenames in os.walk(directory, onerror=onerror):
        for filename in fnmatch.filter(filenames, "[A-Za-z1-9]*.py"):
            mod = import_filename(filename.replace(".py", ""), root)
            if mod is not None:
                result.extend(create_rules(mod, name, modules))

    return result


class ToPy:
    """
    Conversion of a string into Python friendly naming
    """

    def __init__(self, name: str):
        self.name = name
        self.py = name.replace("::", "_").replace("-", "_").lower()
        self.py_class = name.replace("::", "")
        # provider zips has filenames with -
        self.provider = name.replace("::", "-").lower()


class _ObjectEncoder(json.JSONEncoder):
    def default(self, o):
        if hasattr(o, "_value"):
            # pylint: disable=protected-access
            return o._value
        return o


def get_hash(instance: Any) -> str:
    """Return a hash of an object"""
    return hashlib.sha1(
        json.dumps(instance, sort_keys=True, cls=_ObjectEncoder).encode("utf-8")
    ).hexdigest()


def ensure_list(instance: Any) -> list[Any]:
    if isinstance(instance, (list, tuple)):
        return list(instance)
    return [instance]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import regex as re

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.jsonschema._keywords import pattern
from cfnlint.rules import CloudFormationLintRule


class Pattern(CloudFormationLintRule):
    """Check if properties have a valid value"""

    id = "E3031"
    shortdesc = "Check if property values adhere to a specific pattern"
    description = (
        "Check if properties have a valid value in case of a pattern (Regular"
        " Expression)"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#pattern"
    tags = ["resources", "property", "allowed pattern", "regex"]
    child_rules = {
        "W2031": None,
    }

    def __init__(self):
        """Init"""
        super().__init__()
        self.config_definition = {
            "exceptions": {
                "default": [],
                "type": "list",
                "itemtype": "string",
            }
        }
        self.configure()

    def _is_exception(self, instance: str) -> bool:
        for exception in self.config["exceptions"]:
            if re.match(exception, instance):
                return True
        return False

    # pylint: disable=unused-argument, arguments-renamed
    def pattern(
        self, validator: Validator, patrn: str, instance: Any, schema: Any
    ) -> ValidationResult:
        # https://github.com/aws-cloudformation/cfn-lint/issues/3640
        if validator.cfn.has_serverless_transform():
            for _, param in validator.context.parameters.items():
                if param.is_ssm_parameter():
                    if param.ssm_path == instance:
                        return

        if (
            len(validator.context.path.value_path) > 0
            and validator.context.path.value_path[0] == "Parameters"
        ):
            if self.child_rules.get("W2031"):
                yield from self.child_rules["W2031"].pattern(  # type: ignore
                    validator, patrn, instance, schema
                )
            return
        for err in pattern(validator, patrn, instance, schema):
            if not self._is_exception(instance):
                yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ecs.TaskFargateProperties import TaskFargateProperties


@pytest.fixture(scope="module")
def rule():
    rule = TaskFargateProperties()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
            },
            [],
        ),
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": ".25     vCpU",
                "Memory": "512",
            },
            [],
        ),
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 16384,
            },
            [
                ValidationError(
                    "'Memory' is a required property",
                    rule=TaskFargateProperties(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "allOf", 2, "required"]),
                )
            ],
        ),
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Memory": "512",
            },
            [
                ValidationError(
                    "'Cpu' is a required property",
                    rule=TaskFargateProperties(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "allOf", 2, "required"]),
                )
            ],
        ),
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Memory": "512",
                "Cpu": 256,
                "PlacementConstraints": "foo",
            },
            [
                ValidationError(
                    ("'PlacementConstraints' isn't supported for Fargate tasks"),
                    rule=TaskFargateProperties(),
                    path=deque(["PlacementConstraints"]),
                    validator="not",
                    schema_path=deque(["then", "allOf", 2, "not"]),
                )
            ],
        ),
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": {"Ref": "MyParameter"},
                "Memory": "512",
            },
            [],
        ),
        (
            {
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 128,
                "Memory": "512",
            },
            [
                ValidationError(
                    (
                        "128 is not one of ['256', '512', '1024', '2048', "
                        "'4096', '8192', '16384']"
                    ),
                    rule=TaskFargateProperties(),
                    path=deque(["Cpu"]),
                    validator="enum",
                    schema_path=deque(
                        [
                            "then",
                            "allOf",
                            2,
                            "then",
                            "properties",
                            "Cpu",
                            "then",
                            "enum",
                        ]
                    ),
                )
            ],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
                "NetworkMode": {"Ref": "MyParameter"},
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
            },
            [
                ValidationError(
                    ("'NetworkMode' is a required property"),
                    rule=TaskFargateProperties(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "allOf", 0, "required"]),
                )
            ],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
                "NetworkMode": "awsvpc",
                "ContainerDefinitions": [
                    {
                        "LogConfiguration": {
                            "LogDriver": "awslogs",
                            "Options": {
                                "awslogs-group": "log-group",
                                "awslogs-region": "us-east-1",
                                "awslogs-stream-prefix": "ecs",
                            },
                        }
                    }
                ],
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
                "NetworkMode": "awsvpc",
                "ContainerDefinitions": [
                    {
                        "LogConfiguration": {
                            "LogDriver": {"Ref": "MyParameter"},
                            "Options": {
                                "awslogs-group": "log-group",
                                "awslogs-region": "us-east-1",
                                "awslogs-stream-prefix": "ecs",
                            },
                        }
                    }
                ],
            },
            [],
        ),
        (
            {
                "RequiresCompatibilities": ["FARGATE"],
                "Cpu": 256,
                "Memory": "512",
                "NetworkMode": "awsvpc",
                "ContainerDefinitions": [
                    {
                        "LogConfiguration": {
                            "LogDriver": "sumologic",
                            "Options": {
                                "sumo-url": "aurl",
                                "sumo-source-category": "access",
                            },
                        }
                    }
                ],
            },
            [
                ValidationError(
                    ("'sumologic' is not one of ['awslogs', 'splunk', 'awsfirelens']"),
                    rule=TaskFargateProperties(),
                    path=deque(
                        ["ContainerDefinitions", 0, "LogConfiguration", "LogDriver"]
                    ),
                    validator="enum",
                    schema_path=deque(
                        [
                            "then",
                            "allOf",
                            1,
                            "properties",
                            "ContainerDefinitions",
                            "items",
                            "properties",
                            "LogConfiguration",
                            "then",
                            "properties",
                            "LogDriver",
                            "enum",
                        ]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase
from typing import Any, Dict

from cfnlint import ConfigMixIn
from cfnlint.decode.decode import decode_str
from cfnlint.rules import CloudFormationLintRule, Match, RuleMatch, Rules
from cfnlint.runner import TemplateRunner


class TestCloudFormationRuleChild(BaseTestCase):
    """Test CloudFormation Rule"""

    def test_child_rules(self) -> None:
        class TestRuleParent(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]
            child_rules: Dict[str, Any] = {"E1001": None}

            def match(self, _):
                return self.child_rules["E1001"].failure()

        class TestRuleChild(CloudFormationLintRule):
            """Def Rule"""

            id = "E1001"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

            def failure(self):
                return [RuleMatch(["key"], "failure", rule=self)]

        rule_collection = Rules()
        test_rule_parent = TestRuleParent()
        test_rule_child = TestRuleChild()
        rule_collection.register(test_rule_parent)
        rule_collection.register(test_rule_child)

        template, _ = decode_str('{"key": "value"}')
        self.assertIsNotNone(template)
        if template is not None:
            runner = TemplateRunner(None, template, ConfigMixIn([]), rule_collection)
            failures = list(runner.run())

            self.assertListEqual(
                failures,
                [
                    Match(
                        linenumber=1,
                        columnnumber=2,
                        linenumberend=1,
                        columnnumberend=7,
                        filename=None,
                        rule=test_rule_child,
                        message="failure",
                    )
                ],
            )

    def test_child_rules_suppressed(self) -> None:
        class TestRuleParent(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]
            child_rules: Dict[str, Any] = {"E1001": None}

            def match(self, _):
                if self.child_rules.get("E1001"):
                    return self.child_rules["E1001"].failure()
                return []

        class TestRuleChild(CloudFormationLintRule):
            """Def Rule"""

            id = "E1001"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

            def failure(self):
                return [RuleMatch(["key"], "failure", rule=self)]

        rule_collection = Rules()
        test_rule_parent = TestRuleParent()
        test_rule_child = TestRuleChild()
        rule_collection.register(test_rule_parent)
        rule_collection.register(test_rule_child)

        template, _ = decode_str('{"key": "value"}')
        self.assertIsNotNone(template)
        if template is not None:
            runner = TemplateRunner(
                None,
                template,
                ConfigMixIn(ignore_checks=["E1001"]),
                rule_collection,
            )
            failures = list(runner.run())

            self.assertListEqual(failures, [])

    def test_child_rules_configured(self) -> None:
        class TestRuleParent(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]
            child_rules: Dict[str, Any] = {"E1001": None}

            def match(self, _):
                return self.child_rules["E1001"].failure()

        class TestRuleChild(CloudFormationLintRule):
            """Def Rule"""

            id = "E1001"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

            def __init__(self):
                """Init"""
                super(TestRuleChild, self).__init__()
                self.config_definition = {"pass": {"default": True, "type": "boolean"}}
                self.configure()

            def failure(self):
                if self.config["pass"]:
                    return [RuleMatch(["key"], "failure", rule=self)]
                return []

        rule_collection = Rules()
        test_rule_parent = TestRuleParent()
        test_rule_child = TestRuleChild()
        rule_collection.register(test_rule_parent)
        rule_collection.register(test_rule_child)

        template, _ = decode_str('{"key": "value"}')
        self.assertIsNotNone(template)
        if template is not None:
            runner = TemplateRunner(
                None,
                template,
                ConfigMixIn(configure_rules={"E1001": {"pass": False}}),
                rule_collection,
            )
            failures = list(runner.run())

            self.assertListEqual(failures, [])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.formats.LogGroupName import LogGroupName


@pytest.fixture(scope="module")
def rule():
    rule = LogGroupName()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid log group name",
            "123457",
            True,
        ),
        (
            "Valid with invalid type",
            [],
            True,
        ),
        (
            "Valid log group name with special characters",
            "aws/one.two-three#four_five",
            True,
        ),
        ("Invalid log group name because of special character", "foo&bar", False),
        ("Invalid log group name because of empty", "", False),
        ("Invalid log group name because too long", "a" * 513, False),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    result = rule.format(validator, instance)
    assert result == expected, f"Test {name!r} got {result!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context._mappings import Map, Mappings, _MappingSecondaryKey


@pytest.mark.parametrize(
    "name,get_key_1,get_key_2,expected",
    [
        ("Valid get", "A", "B", ["C"]),
        ("Invalid first key", "B", "B", KeyError()),
        ("Invalid second key", "A", "C", KeyError()),
    ],
)
def test_mapping_value(name, get_key_1, get_key_2, expected):
    mapping = Map.create_from_dict({"A": {"B": "C"}})

    if isinstance(expected, Exception):
        with pytest.raises(type(expected)):
            list(mapping.find_in_map(get_key_1, get_key_2))
    else:
        assert list(mapping.find_in_map(get_key_1, get_key_2)) == expected


def test_transforms():
    mapping = Map.create_from_dict({"A": {"Fn::Transform": "C"}})

    assert mapping.keys.get("A").is_transform is True

    mapping = Map.create_from_dict({"Fn::Transform": {"B": "C"}})

    assert mapping.is_transform is True

    mapping = Mappings.create_from_dict({"Fn::Transform": {"B": {"C": "D"}}})

    assert mapping.is_transform is True


@pytest.mark.parametrize(
    "name,mappings,expected",
    [
        (
            "Valid mappings",
            {
                "A": {"B": {"C": "D"}},
                "1": {"2": {"3": "4"}},
                "Z": [],
                "9": {"8": []},
                "M": {"N": {"O": {"P": "Q"}}},
            },
            Mappings(
                {
                    "A": Map({"B": _MappingSecondaryKey({"C": "D"})}),
                    "1": Map({"2": _MappingSecondaryKey({"3": "4"})}),
                    "Z": Map({}),
                    "9": Map({"8": _MappingSecondaryKey({})}),
                    "M": Map({"N": _MappingSecondaryKey({})}),
                }
            ),
        ),
        (
            "Valid mappings with transforms",
            {
                "A": {"Fn::Transform": "MyTransform"},
                "1": {"2": {"Fn::Transform": "MyTransform"}},
            },
            Mappings(
                {
                    "A": Map({}, True),
                    "1": Map({"2": _MappingSecondaryKey({}, True)}),
                }
            ),
        ),
        (
            "Valid mappings with transforms for mappings",
            {
                "Fn::Transform": "MyTransform",
            },
            Mappings({}, True),
        ),
        (
            "Invalid mappings with wrong types",
            {
                "A": {True: {"C": "foo"}},
                "1": {"2": {False: "foo"}},
            },
            Mappings(
                {
                    "A": Map({}, False),
                    "1": Map({"2": _MappingSecondaryKey({}, False)}),
                }
            ),
        ),
    ],
)
def test_mapping_creation(name, mappings, expected):
    results = Mappings.create_from_dict(mappings)

    assert results == expected, f"{name!r} failed got {results!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.GetAttFormat import GetAttFormat


@pytest.fixture(scope="module")
def rule():
    rule = GetAttFormat()
    rule._exceptions = ["AWS::EC2::SecurityGroup.Id"]
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyBucket": {"Type": "AWS::S3::Bucket"},
            "MyVpc": {"Type": "AWS::EC2::VPC"},
            "MySecurityGroup": {"Type": "AWS::EC2::SecurityGroup"},
            "MyCustomResource": {"Type": "Custom::CustomResource"},
            "MySubTemplate": {"Type": "AWS::CloudFormation::Stack"},
            "MyProvisionedProduct": {
                "Type": "AWS::ServiceCatalog::CloudFormationProvisionedProduct"
            },
            "MySSMParameter": {"Type": "AWS::SSM::Parameter"},
        },
    }


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid GetAtt with a good format",
            ["MyVpc", "VpcId"],
            {"format": "AWS::EC2::VPC.Id"},
            [],
        ),
        (
            "Valid GetAtt to a custom resource",
            ["MyCustomResource", "ImageId"],
            {"format": "AWS::EC2::VPC.Id"},
            [],
        ),
        (
            "Valid GetAtt to a sub template ",
            ["MySubTemplate", "Outputs.ImageId"],
            {"format": "AWS::EC2::VPC.Id"},
            [],
        ),
        (
            "Valid GetAtt to a provisioned product ",
            ["MyProvisionedProduct", "Outputs.VpcId"],
            {"format": "AWS::EC2::VPC.Id"},
            [],
        ),
        (
            "Valid GetAtt because of exception",
            ["MyBucket", "Arn"],
            {"format": "AWS::EC2::SecurityGroup.Id"},
            [],
        ),
        (
            "Valid GetAtt because of exception with attribute",
            ["MySSMParameter", "Value"],
            {"format": "AWS::EC2::Image.Id"},
            [],
        ),
        (
            "Invalid GetAtt with a bad format",
            ["MyBucket", "Arn"],
            {"format": "AWS::EC2::VPC.Id"},
            [
                ValidationError(
                    (
                        "{'Fn::GetAtt': ['MyBucket', 'Arn']} "
                        "does not match destination format of 'AWS::EC2::VPC.Id'"
                    ),
                    rule=GetAttFormat(),
                )
            ],
        ),
        (
            "Invalid GetAtt with a getatt and no format",
            ["MyBucket", "WebsiteURL"],
            {"format": "AWS::EC2::VPC.Id"},
            [
                ValidationError(
                    (
                        "{'Fn::GetAtt': ['MyBucket', 'WebsiteURL']} "
                        "with formats ['uri'] does not match "
                        "destination format of 'AWS::EC2::VPC.Id'"
                    ),
                    rule=GetAttFormat(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.validate(validator, schema, instance, schema))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

from cfnlint.helpers import ensure_list, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.helpers import get_resource_by_name, get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class ServiceFargate(CfnLintKeyword):
    id = "E3054"
    shortdesc = (
        "Validate ECS service using Fargate uses TaskDefinition that allows Fargate"
    )
    description = (
        "When using an ECS service with 'LaunchType' of 'FARGATE' "
        "the associated task definition must have 'RequiresCompatibilities' "
        "specified with 'FARGATE' listed"
    )
    tags = ["resources", "ecs"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::Service/Properties"],
        )

    def _filter_resource_name(self, instance: Any) -> str | None:
        fn_k, fn_v = is_function(instance)
        if fn_k is None:
            return None
        if fn_k == "Ref":
            if isinstance(fn_v, str):
                return fn_v
        elif fn_k == "Fn::GetAtt":
            name = ensure_list(fn_v)[0].split(".")[0]
            if isinstance(name, str):
                return name
        return None

    def _get_service_properties(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[str, str, Validator]]:
        for task_definition_id, task_definition_validator in get_value_from_path(
            validator, instance, deque(["TaskDefinition"])
        ):
            task_definition_resource_name = self._filter_resource_name(
                task_definition_id
            )
            if task_definition_resource_name is None:
                continue

            for (
                launch_type,
                launch_type_validator,
            ) in get_value_from_path(
                task_definition_validator, instance, deque(["LaunchType"])
            ):
                yield (
                    task_definition_resource_name,
                    launch_type,
                    launch_type_validator,
                )

    def _get_task_definition_properties(
        self, validator: Validator, resource_name: Any
    ) -> Iterator[tuple[list[Any] | None, Validator]]:
        task_definition, task_definition_validator = get_resource_by_name(
            validator, resource_name, ["AWS::ECS::TaskDefinition"]
        )
        if not task_definition:
            return
        for capabilities, capabilities_validator in get_value_from_path(
            task_definition_validator,
            task_definition,
            path=deque(["Properties", "RequiresCompatibilities"]),
        ):
            for network_mode, network_mode_validator in get_value_from_path(
                capabilities_validator,
                task_definition,
                path=deque(["Properties", "NetworkMode"]),
            ):

                if network_mode == "awsvpc" or network_mode_validator.is_type(
                    network_mode, "object"
                ):
                    continue
                if capabilities is None:
                    yield capabilities, capabilities_validator
                    continue
                if not isinstance(capabilities, list):
                    continue
                for capibility, _ in get_value_from_path(
                    network_mode_validator,
                    capabilities,
                    path=deque(["*"]),
                ):
                    if isinstance(capibility, dict) or capibility == "FARGATE":
                        break
                else:
                    yield capabilities, capabilities_validator

    def validate(
        self, validator: Validator, _: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for (
            task_definition_resource_name,
            launch_type,
            service_validator,
        ) in self._get_service_properties(
            validator,
            instance,
        ):
            if launch_type != "FARGATE":
                continue
            for (
                capabilities,
                capabilities_validator,
            ) in self._get_task_definition_properties(
                service_validator,
                task_definition_resource_name,
            ):
                if capabilities is None:
                    yield ValidationError(
                        "'RequiresCompatibilities' is a required property",
                        validator="required",
                        rule=self,
                        path_override=deque(
                            list(capabilities_validator.context.path.path)[:-1]
                        ),
                    )
                    continue

                yield ValidationError(
                    f"{capabilities!r} does not contain items matching 'FARGATE'",
                    validator="contains",
                    rule=self,
                    path_override=capabilities_validator.context.path.path,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.templates.ApproachingLimitDescription import (
    ApproachingLimitDescription,
)


@pytest.fixture(scope="module")
def rule():
    rule = ApproachingLimitDescription()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid description",
            "a" * 18,
            [],
        ),
        (
            "Too long",
            "a" * 19,
            [
                ValidationError(
                    f"'{'a'*19}' is approaching the max length of 20",
                    rule=ApproachingLimitDescription(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errors = list(rule.maxLength(validator, 20, instance, {}))

    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.VpcSubnetCidr import VpcSubnetCidr


@pytest.fixture(scope="module")
def rule():
    rule = VpcSubnetCidr()
    yield rule


_template = {
    "Resources": {
        "Vpc": {
            "Type": "AWS::EC2::VPC",
        },
        "VpcV4Subnet1": {
            "Type": "AWS::EC2::Subnet",
            "Properties": {
                "CidrBlock": "10.0.0.0/24",
                "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
            },
        },
        "SecurityGroupIngress": {
            "Type": "AWS::EC2::SecurityGroupIngress",
            "Properties": {
                "VpcId": {"Fn::GetAtt": ["Vpc", "DefaultSecurityGroup"]},
            },
        },
    }
}


@pytest.mark.parametrize(
    "name,instance,template,path,expected",
    [
        (
            "Valid with no Private Ip Address",
            {
                "CidrBlock": "10.0.0.0/16",
            },
            _template,
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Valid with bad ip address on VPC",
            {
                "CidrBlock": "10.0.0/16",
            },
            _template,
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Valid with a bad subnet IP Address",
            {
                "CidrBlock": "10.0.0.0/16",
            },
            {
                "Parameters": {"VpcCidr": {"Type": "String"}},
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "CidrBlock": "10.0.0/24",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Parameter CIDR",
            {
                "CidrBlock": {"Ref": "VpcCidr"},
            },
            {
                "Parameters": {"VpcCidr": {"Type": "String"}},
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "CidrBlock": "10.0.0.0/24",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Empty CidrBlocks should continue",
            {},
            _template,
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Subnet outside of the VPC",
            {
                "CidrBlock": "11.0.0.0/16",
            },
            _template,
            {"path": ["Resources", "Vpc", "Properties"]},
            [
                ValidationError(
                    "'10.0.0.0/24' is not a valid subnet of ['11.0.0.0/16']",
                    rule=VpcSubnetCidr(),
                    path_override=deque(
                        ["Resources", "VpcV4Subnet1", "Properties", "CidrBlock"]
                    ),
                )
            ],
        ),
        (
            "Subnet larger than VPC",
            {
                "CidrBlock": "10.0.0.0/32",
            },
            _template,
            {"path": ["Resources", "Vpc", "Properties"]},
            [
                ValidationError(
                    "'10.0.0.0/24' is not a valid subnet of ['10.0.0.0/32']",
                    rule=VpcSubnetCidr(),
                    path_override=deque(
                        ["Resources", "VpcV4Subnet1", "Properties", "CidrBlock"]
                    ),
                )
            ],
        ),
        (
            "Subnet IPV6 CIDR with no ipv6 allocations",
            {
                "CidrBlock": "11.0.0.0/16",
            },
            {
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "Ipv6CidrBlock": "2001:db8::/32",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [
                ValidationError(
                    (
                        "'2001:db8::/32' is specified on a VPC that "
                        "has no ipv6 networks defined"
                    ),
                    rule=VpcSubnetCidr(),
                    path_override=deque(
                        ["Resources", "VpcV4Subnet1", "Properties", "Ipv6CidrBlock"]
                    ),
                )
            ],
        ),
        (
            "Subnet IPV6 CIDR valid with addtional block",
            {"CidrBlock": "10.0.0.0/16"},
            {
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcCidr": {
                        "Type": "AWS::EC2::VPCCidrBlock",
                        "Properties": {
                            "Ipv6CidrBlock": "fc00::/7",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "Ipv6CidrBlock": "fc00::/16",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Invalid subnet IPV6 CIDR valid with addtional block",
            {"CidrBlock": "10.0.0.0/32"},
            {
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcCidr": {
                        "Type": "AWS::EC2::VPCCidrBlock",
                        "Properties": {
                            "Ipv6CidrBlock": "fc00::/32",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "Ipv6CidrBlock": "fc00::/16",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [
                ValidationError(
                    "'fc00::/16' is not a valid subnet of ['fc00::/32']",
                    rule=VpcSubnetCidr(),
                    path_override=deque(
                        ["Resources", "VpcV4Subnet1", "Properties", "Ipv6CidrBlock"]
                    ),
                )
            ],
        ),
        (
            "Subnet IPV4 CIDR valid with IPam Pool ID",
            {"Ipv4IpamPoolId": "poolid"},
            _template,
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Subnet IPV4 CIDR valid with multiple CIDRs",
            {"CidrBlock": "10.0.0.0/16"},
            {
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcCidr": {
                        "Type": "AWS::EC2::VPCCidrBlock",
                        "Properties": {
                            "CidrBlock": "11.0.0.0/16",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "CidrBlock": "11.0.0.0/24",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Subnet IPV4 CIDR invalid with multiple CIDRs with ipam pool ID",
            {"CidrBlock": "10.0.0.0/16"},
            {
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcCidr": {
                        "Type": "AWS::EC2::VPCCidrBlock",
                        "Properties": {
                            "Ipv4IpamPoolId": "pool-id",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "CidrBlock": "11.0.0.0/16",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
        (
            "Subnet IPV6 CIDR invalid with multiple CIDRs and one ipam pool ID",
            {"CidrBlock": "10.0.0.0/16"},
            {
                "Resources": {
                    "Vpc": {
                        "Type": "AWS::EC2::VPC",
                    },
                    "VpcCidr1": {
                        "Type": "AWS::EC2::VPCCidrBlock",
                        "Properties": {
                            "Ipv6IpamPoolId": "pool-id",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                    "VpcCidr2": {
                        "Type": "AWS::EC2::VPCCidrBlock",
                        "Properties": {
                            "Ipv6CidrBlock": "fc00::/32",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                    "VpcV4Subnet1": {
                        "Type": "AWS::EC2::Subnet",
                        "Properties": {
                            "Ipv6CidrBlock": "fc00::/16",
                            "VpcId": {"Fn::GetAtt": ["Vpc", "VpcId"]},
                        },
                    },
                },
            },
            {"path": ["Resources", "Vpc", "Properties"]},
            [],
        ),
    ],
    indirect=["template", "path"],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert (
        errs == expected
    ), f"Expected test {name!r} to have {expected!r} but got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any, Sequence

import cfnlint.data.schemas.extensions.aws_ecs_taskdefinition
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class LogConfiguration(CfnLintJsonSchema):
    id = "E3046"
    shortdesc = "Validate ECS task logging configuration for awslogs"
    description = (
        "When 'awslogs' the options 'awslogs-group' and 'awslogs-region' are required"
    )
    tags = ["resources"]

    def __init__(
        self, keywords: Sequence[str] | None = None, all_matches: bool = False
    ) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ECS::TaskDefinition/Properties/ContainerDefinitions/*/LogConfiguration"
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ecs_taskdefinition,
                filename="logging_configuration.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        # if the schema has a description will only replace the message with that
        # description and use the best error for the location information
        if not self._use_schema_arg:
            schema = self._schema

        cfn_validator = self.extend_validator(
            validator=validator.evolve(
                function_filter=validator.function_filter.evolve(
                    validate_dynamic_references=False,
                    add_cfn_lint_keyword=False,
                )
            ),
            schema=schema,
            context=validator.context.evolve(),
        )

        yield from self._iter_errors(cfn_validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.GetAtt import GetAtt
from cfnlint.rules.functions.GetAz import GetAz
from cfnlint.rules.functions.Ref import Ref
from cfnlint.rules.functions.Sub import Sub
from cfnlint.rules.functions.SubResolved import SubResolved
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = Sub()
    rule.child_rules["W1031"] = SubResolved()
    yield rule


@pytest.fixture
def cfn():
    return Template(
        "",
        {
            "Parameters": {
                "MyParameter": {
                    "Type": "String",
                    "AllowedValues": [
                        "one",
                        "two",
                    ],
                }
            },
            "Resources": {
                "MyResource": {"Type": "AWS::S3::Bucket"},
                "MySimpleAd": {"Type": "AWS::DirectoryService::SimpleAD"},
                "MyPolicy": {"Type": "AWS::IAM::ManagedPolicy"},
            },
        },
        regions=["us-east-1"],
    )


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Sub",
            {"Fn::Sub": "foo"},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Sub with an incorrect type",
            {"Fn::Sub": {"foo": "bar"}},
            {"type": "string"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'array', 'string'",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with an invalid output type",
            {"Fn::Sub": "foo"},
            {"type": "array"},
            [
                ValidationError(
                    "{'Fn::Sub': 'foo'} is not of type 'array'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Valid Fn::Sub with a valid Ref",
            {"Fn::Sub": "${AWS::Region}"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with a valid Ref and a string escape",
            {"Fn::Sub": "${!Foo}=${AWS::Region}"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with a valid Ref and a string escape with parameter",
            {"Fn::Sub": "${!foo}=${bar}"},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "'bar' is not one of ["
                        "'MyParameter', 'MyResource', 'MySimpleAd', 'MyPolicy', "
                        "'AWS::AccountId', 'AWS::NoValue', 'AWS::NotificationARNs', "
                        "'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a invalid Ref",
            {"Fn::Sub": "${foo}"},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "'foo' is not one of ["
                        "'MyParameter', 'MyResource', 'MySimpleAd', 'MyPolicy', "
                        "'AWS::AccountId', 'AWS::NoValue', 'AWS::NotificationARNs', "
                        "'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Valid Fn::Sub with a Ref to a sub parameter",
            {"Fn::Sub": ["${foo}", {"foo": "bar"}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with an escape character",
            {"Fn::Sub": "${!foo}"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with an ref to resource",
            {"Fn::Sub": ["${MyResource}", {"MyResource": {"Ref": "MyResource"}}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with an integer parameter",
            {"Fn::Sub": ["${Foo}", {"Foo": 1}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with a space and escape char",
            {"Fn::Sub": "${ !Foo }"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with a space and no escape char",
            {"Fn::Sub": "${ AWS::AccountId }"},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Sub with a function",
            {"Fn::Sub": {"Fn::Split": [",", "${AWS::AccountId}"]}},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "{'Fn::Split': [',', '${AWS::AccountId}']} is "
                        "not of type 'array', 'string'"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a too to many elements",
            {"Fn::Sub": ["${foo}", {"foo": "bar"}, {}]},
            {"type": "string"},
            [
                ValidationError(
                    "expected maximum item count: 2, found: 3",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["cfnContext", "schema", "maxItems"]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a bad object",
            {"Fn::Sub": ["${foo}", []]},
            {"type": "string"},
            [
                ValidationError(
                    "[] is not of type 'object'",
                    path=deque(["Fn::Sub", 1]),
                    schema_path=deque(
                        ["cfnContext", "schema", "prefixItems", 1, "type"]
                    ),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a bad object type",
            {
                "Fn::Sub": [
                    "${foo}",
                    {
                        "foo": [],
                    },
                ]
            },
            {"type": "string"},
            [
                ValidationError(
                    "[] is not of type 'string'",
                    path=deque(["Fn::Sub", 1, "foo"]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            1,
                            "patternProperties",
                            "[a-zA-Z0-9]+",
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a bad object type from fn",
            {
                "Fn::Sub": [
                    "${foo}",
                    {
                        "foo": {"Fn::GetAZs": ""},
                    },
                ]
            },
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::GetAZs': ''} is not of type 'string'",
                    path=deque(["Fn::Sub", 1, "foo"]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            1,
                            "patternProperties",
                            "[a-zA-Z0-9]+",
                            "cfnContext",
                            "schema",
                            "fn_getazs",
                        ]
                    ),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Valid Fn::Sub with a GetAtt",
            {"Fn::Sub": "${MyResource.Arn}"},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Sub with a GetAtt and a bad attribute",
            {"Fn::Sub": "${MyResource.Foo}"},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "'Foo' is not one of ['Arn', "
                        "'DomainName', "
                        "'DualStackDomainName', "
                        "'RegionalDomainName', "
                        "'MetadataTableConfiguration.S3TablesDestination.TableNamespace',"
                        " 'MetadataTableConfiguration.S3TablesDestination.TableArn',"
                        " 'WebsiteURL'] in ['us-east-1']"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a GetAtt and a bad resource name",
            {"Fn::Sub": "${Foo.Bar}"},
            {"type": "string"},
            [
                ValidationError(
                    "'Foo' is not one of ['MyResource', 'MySimpleAd', 'MyPolicy']",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a GetAtt to an array of attributes",
            {"Fn::Sub": "${MySimpleAd.DnsIpAddresses}"},
            {"type": "string"},
            [
                ValidationError(
                    ("'MySimpleAd.DnsIpAddresses' is not of type 'string'"),
                    instance="MySimpleAd.DnsIpAddresses",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a GetAtt to an integer",
            {"Fn::Sub": "${MyPolicy.AttachmentCount}"},
            {"type": "string"},
            [
                ValidationError(
                    ("'MyPolicy.AttachmentCount' is not of type 'string'"),
                    instance="MySimpleAd.DnsIpAddresses",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "One valid resolution",
            {"Fn::Sub": "${MyParameter}"},
            {"type": "string", "const": "two"},
            [],
        ),
        (
            "No valid resolution",
            {"Fn::Sub": "${MyParameter}"},
            {"type": "string", "const": "three"},
            [
                ValidationError(
                    ("'three' was expected when 'Fn::Sub' is resolved"),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["const"]),
                    validator="fn_sub",
                    rule=SubResolved(),
                ),
                ValidationError(
                    ("'three' was expected when 'Fn::Sub' is resolved"),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["const"]),
                    validator="fn_sub",
                    rule=SubResolved(),
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, context, cfn):
    validator = CfnTemplateValidator({}).extend(
        validators={
            "ref": Ref().ref,
            "fn_getatt": GetAtt().fn_getatt,
            "fn_getazs": GetAz().fn_getazs,
        }
    )(context=context.evolve(strict_types=False), cfn=cfn)
    errs = list(rule.fn_sub(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

from cfnlint.helpers import ensure_list, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.helpers import get_resource_by_name, get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class ServiceNetworkConfiguration(CfnLintKeyword):
    id = "E3052"
    shortdesc = "Validate ECS service requires NetworkConfiguration"
    description = (
        "When using an ECS task definition has NetworkMode set to "
        "'awsvpc' then 'NetworkConfiguration' is required"
    )
    tags = ["resources", "ecs"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::Service/Properties"],
        )

    def _filter_resource_name(self, instance: Any) -> str | None:
        fn_k, fn_v = is_function(instance)
        if fn_k is None:
            return None
        if fn_k == "Ref":
            if isinstance(fn_v, str):
                return fn_v
        elif fn_k == "Fn::GetAtt":
            name = ensure_list(fn_v)[0].split(".")[0]
            if isinstance(name, str):
                return name
        return None

    def _get_service_properties(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[str, str, Validator]]:
        for task_definition_id, task_definition_validator in get_value_from_path(
            validator, instance, deque(["TaskDefinition"])
        ):
            task_definition_resource_name = self._filter_resource_name(
                task_definition_id
            )
            if task_definition_resource_name is None:
                continue

            for (
                network_configuration,
                network_configuration_validator,
            ) in get_value_from_path(
                task_definition_validator, instance, deque(["NetworkConfiguration"])
            ):
                yield (
                    task_definition_resource_name,
                    network_configuration,
                    network_configuration_validator,
                )

    def _get_task_definition_properties(
        self, validator: Validator, resource_name: Any
    ) -> Iterator[tuple[str | int | None, Validator]]:
        target_group, target_group_validator = get_resource_by_name(
            validator, resource_name, ["AWS::ECS::TaskDefinition"]
        )
        if not target_group:
            return

        for network_mode, network_mode_validator in get_value_from_path(
            target_group_validator,
            target_group,
            path=deque(["Properties", "NetworkMode"]),
        ):
            if network_mode == "awsvpc":
                yield network_mode, network_mode_validator

    def validate(
        self, validator: Validator, _: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for (
            task_definition_resource_name,
            network_configuration,
            task_definition_validator,
        ) in self._get_service_properties(
            validator,
            instance,
        ):
            for _, _ in self._get_task_definition_properties(
                task_definition_validator,
                task_definition_resource_name,
            ):
                if network_configuration is None:
                    yield ValidationError(
                        "'NetworkConfiguration' is a required property",
                        validator="required",
                        rule=self,
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import TRANSFORM_LANGUAGE_EXTENSION
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class ToJsonString(BaseFn):
    """Check if ToJsonString values are correct"""

    id = "E1031"
    shortdesc = "ToJsonString validation of parameters"
    description = "Making sure Fn::ToJsonString is configured correctly"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html"
    tags = ["functions", "toJsonString"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::ToJsonString",
            ("string",),
            resolved_rule="W1040",
        )

    def fn_tojsonstring(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.context.transforms.has_language_extensions_transform():
            yield ValidationError(
                (
                    f"{self.fn.name} is not supported without "
                    f"{TRANSFORM_LANGUAGE_EXTENSION!r} transform"
                ),
                validator=self.fn.py,
                rule=self,
            )
            return

        for err in super().validate(validator, s, instance, schema):
            err.rule = self
            yield err
